from suzieq.gui.guiutils import display_help_icon, display_title
from suzieq.gui.guiutils import (gui_get_df, get_base_url, get_session_id,
                                 SuzieqMainPages)
from suzieq.sqobjects.path import PathObj
from copy import copy
from urllib.parse import quote
from typing import Tuple
import graphviz as graphviz
import pandas as pd
import streamlit as st
from dataclasses import dataclass, asdict, field


def get_title():
    return 'Path'


def make_fields_failed_df():
    '''return a dictionary as this is the only way dataclassses work'''

    return [
        {'name': 'device', 'df': pd.DataFrame(), 'query': 'status == "dead"'},
        {'name': 'interfaces', 'df': pd.DataFrame(), 'query': 'state == "down"'},
        {'name': 'mlag', 'df': pd.DataFrame(),
         'query': 'mlagSinglePortsCnt != 0 or mlagErrorPortsCnt != 0'},
        {'name': 'ospf', 'df': pd.DataFrame(), 'query': 'adjState == "other"'},
        {'name': 'bgp', 'df': pd.DataFrame(), 'query': 'state == "NotEstd"'}
    ]


@dataclass
class PathSessionState:
    page: str = SuzieqMainPages.PATH
    namespace: str = ''
    source: str = ''
    dest: str = ''
    start_time: str = ''
    end_time: str = ''
    show_ifnames: bool = False
    vrf: str = ''
    pathobj: PathObj = None
    path_df: pd.DataFrame = None


@dataclass
class FailedDFs:
    dfs: list = field(default_factory=make_fields_failed_df)


def get_path_url_params(state: PathSessionState):
    '''Return the list of valid query params for path URL'''
    state_dict = copy(asdict(state))
    del state_dict['pathobj']
    del state_dict['path_df']

    return state_dict


def gui_path_summarize(path_df: pd.DataFrame) -> pd.DataFrame:
    '''Summarizes the path, a copy of the function in pandas/path.py'''

    if path_df.empty:
        return pd.DataFrame()

    namespace = path_df.namespace.iloc[0]
    ns = {}
    ns[namespace] = {}

    perhopEcmp = path_df.query('hopCount != 0') \
                        .groupby(by=['hopCount'])['hostname']
    ns[namespace]['totalPaths'] = path_df['pathid'].max()
    ns[namespace]['perHopEcmp'] = perhopEcmp.nunique().tolist()
    ns[namespace]['maxPathLength'] = path_df.groupby(by=['pathid'])[
        'hopCount'].max().max()
    ns[namespace]['avgPathLength'] = path_df.groupby(by=['pathid'])[
        'hopCount'].max().mean()
    ns[namespace]['uniqueDevices'] = path_df['hostname'].nunique()
    ns[namespace]['mtuMismatch'] = not all(path_df['mtuMatch'])
    ns[namespace]['usesOverlay'] = any(path_df['overlay'])
    ns[namespace]['pathMtu'] = min(path_df.query('iif != "lo"')['inMtu'].min(),
                                   path_df.query('iif != "lo"')['outMtu'].min())

    summary_fields = ['totalPaths', 'perHopEcmp', 'maxPathLength',
                      'avgPathLength', 'uniqueDevices', 'pathMtu',
                      'usesOverlay', 'mtuMismatch']
    return pd.DataFrame(ns).reindex(summary_fields, axis=0) \
                           .convert_dtypes()


def highlight_erroneous_rows(row):
    '''Highlight rows with error in them'''
    if getattr(row, 'error', '') != '':
        return [f"background-color: red; color: white"]*len(row)
    else:
        return [""]*len(row)


@st.cache(ttl=120, allow_output_mutation=True, show_spinner=False,
          max_entries=10)
def path_get(state: PathSessionState, pathobj: PathObj,
             forward_dir: bool) -> Tuple[pd.DataFrame, pd.DataFrame]:
    '''Run the path and return the dataframes'''
    if forward_dir:
        df = pathobj.get(namespace=[state.namespace], source=state.source,
                         dest=state.dest, vrf=state.vrf)

        summ_df = gui_path_summarize(df)
    return df, summ_df


@st.cache(ttl=120, allow_output_mutation=True, show_spinner=False,
          hash_funcs={"streamlit.delta_generator.DeltaGenerator": lambda x: 1},
          max_entries=10)
def get_failed_data(namespace: str, pgbar, sqobjs) -> FailedDFs:
    '''Get interface/mlag/routing protocol states that are failed'''

    faileddfs = FailedDFs()

    progress = 40
    for i, entry in enumerate(faileddfs.dfs):
        entry['df'] = gui_get_df(sqobjs[entry['name']],
                                 namespace=[namespace])
        if not entry['df'].empty and (entry.get('query', '')):
            entry['df'] = entry['df'].query(entry['query'])

        pgbar.progress(progress + i*10)

    return faileddfs


def path_sidebar(state, sqobjs):
    """Configure sidebar"""

    devdf = gui_get_df(sqobjs['device'], columns=['namespace'])
    if devdf.empty:
        st.error('Unable to retrieve any namespace info')
        st.stop()

    namespaces = sorted(devdf.namespace.unique().tolist())
    if state.namespace:
        nsidx = namespaces.index(state.namespace)
    else:
        nsidx = 0

    url = '&amp;'.join([
        f'{get_base_url()}?page=_Help_',
        'help=yes',
        'help_on=Path',
    ])
    display_help_icon(url)
    with st.sidebar:
        with st.form(key='trace'):

            state.namespace = st.selectbox('Namespace',
                                           namespaces, key='path_namespace', index=nsidx)
            src_ph = st.empty()
            dst_ph = st.empty()
            state.source = src_ph.text_input('Source IP', key='path_source',
                                             value=state.source)
            state.dest = dst_ph.text_input(
                'Dest IP', key='path_dest', value=state.dest)

            state.vrf = st.text_input('VRF', value=state.vrf,
                                      key='path_vrf')
            state.start_time = st.text_input('Start Time',
                                             value=state.start_time,
                                             key='path_start_time')
            state.end_time = st.text_input('End Time',
                                           value=state.end_time,
                                           key='path_end_time')

            submit = st.form_submit_button('Trace', on_click=path_sync_state)

        state.show_ifnames = st.checkbox('Show in/out interface names',
                                         value=state.show_ifnames,
                                         key='path_show_ifnames',
                                         on_change=path_sync_state)
        swap_src_dest = st.button(
            'Source <-> Dest', key='path_swap', on_click=path_sync_state)
        if swap_src_dest:
            source, dest = state.dest, state.source
            source = src_ph.text_input('Source IP', value=source)
            dest = dst_ph.text_input('Dest IP', value=dest)
            state.source, state.dest = state.dest, state.source
    return


def get_node_tooltip_color(hostname: str, faileddfs: FailedDFs) -> str:
    '''Get tooltip and node color for node based on values in various tables'''

    ttip = {'title': ['Failed entry count']}
    for entry in faileddfs.dfs:
        mdf = entry['df']
        ttip.update({entry['name']: [mdf.loc[mdf.hostname == hostname]
                                     .hostname.count()]})
    tdf = pd.DataFrame(ttip)
    if tdf.select_dtypes(include='int64').max().max() > 0:
        color = 'red'
    else:
        color = 'black'
    return('\n'.join(tdf.T.to_string().split('\n')[1:]), color)


@st.cache(max_entries=10, allow_output_mutation=True)
def build_graphviz_obj(show_ifnames: bool, df: pd.DataFrame,
                       faileddfs: FailedDFs):
    '''Return a graphviz object'''

    graph_attr = {'splines': 'polyline', 'layout': 'dot'}
    if show_ifnames:
        graph_attr.update({'nodesep': '1.0'})

    g = graphviz.Digraph(graph_attr=graph_attr,
                         name='Hover over arrow head for edge info')

    hostset = set()
    for hostgroup in df.groupby(by=['hopCount']).hostname.unique().tolist():
        with g.subgraph() as s:
            s.attr(rank='same')
            for hostname in hostgroup:
                if hostname in hostset:
                    continue
                hostset.add(hostname)
                debugURL = '&amp;'.join([
                    f'{get_base_url()}?page={quote("_Path_Debug_")}',
                    'lookupType=hop',
                    f'namespace={quote(df.namespace.unique().tolist()[0])}',
                    f'session={quote(get_session_id())}',
                    f'hostname={quote(hostname)}',
                ])
                tooltip, color = get_node_tooltip_color(hostname, faileddfs)
                s.node(hostname, tooltip=tooltip, color=color, URL=debugURL,
                       target='_graphviz', shape='box')

    pathid = 0
    prevrow = None
    connected_set = set()

    df['nextPathid'] = df.pathid.shift(-1).fillna('0').astype(int)
    for row in df.itertuples():
        if row.pathid != pathid:
            prevrow = row
            pathid = row.pathid
            continue
        conn = (prevrow.hostname, row.hostname)
        if conn not in connected_set:
            if row.overlay:
                path_type = 'underlay'
                color = 'purple'
            elif prevrow.isL2:
                path_type = 'l2'
                color = 'blue'
            else:
                path_type = 'l3'
                color = 'black'

            if not row.mtuMatch:
                color = 'red'
                error = 'MTU mismatch'
                err_pfx = ', '
            else:
                error = ''
                err_pfx = ''

            tdf = pd.DataFrame({
                'pathType': path_type,
                'protocol': [prevrow.protocol],
                'ipLookup': [prevrow.ipLookup],
                'vtepLookup': [prevrow.vtepLookup],
                'macLookup': [prevrow.macLookup],
                'nexthopIp': [prevrow.nexthopIp],
                'vrf': [prevrow.vrf],
                'mtu': [f'{prevrow.outMtu} -> {row.inMtu}'],
                'oif': [prevrow.oif],
                'iif': [row.iif]})
            rowerr = getattr(prevrow, 'error', '')
            if rowerr:
                error += f"{err_pfx}{rowerr}"
            if error:
                err_pfx = ', '
            if row.nextPathid != row.pathid:
                # We need to capture any errors on the dest node as well
                destnode_error = getattr(row, 'error', '')
                if destnode_error:
                    error += f'{err_pfx}{destnode_error}'

            if error:
                tdf['error'] = error
                color = 'red'
            tooltip = '\n'.join(tdf.T.to_string(
                justify='right').split('\n')[1:])
            debugURL = '&amp;'.join([
                f'{get_base_url()}?page={quote("_Path_Debug_")}',
                'lookupType=edge',
                f'namespace={quote(row.namespace)}',
                f'session={quote(get_session_id())}',
                f'hostname={quote(prevrow.hostname)}',
                f'vrf={quote(prevrow.vrf)}',
                f'vtepLookup-{prevrow.vtepLookup}',
                f'ifhost={quote(row.hostname)}',
                f'ipLookup={quote(prevrow.ipLookup)}',
                f'oif={quote(prevrow.oif)}',
                f'macaddr={quote(prevrow.macLookup or "")}',
                f'nhip={quote(prevrow.nexthopIp)}',
            ])
            if show_ifnames:
                g.edge(prevrow.hostname, row.hostname, color=color,
                       label=str(row.hopCount), URL=debugURL,
                       edgetarget='_graphviz', tooltip=tooltip,
                       taillabel=prevrow.oif, headlabel=row.iif,
                       penwidth='2.0',
                       )
            else:
                g.edge(prevrow.hostname, row.hostname, color=color,
                       label=str(row.hopCount), URL=debugURL,
                       edgetarget='_graphviz', penwidth='2.0',
                       tooltip=tooltip
                       )

            connected_set.add(conn)
        prevrow = row
    df.drop(columns=['nextPathid'], inplace=True, errors='ignore')
    return g


def path_sync_state():
    wsstate = st.session_state
    state = wsstate.pathSessionState

    state.namespace = wsstate.path_namespace
    state.source = wsstate.path_source
    state.dest = wsstate.path_dest
    state.vrf = wsstate.path_vrf
    state.start_time = wsstate.path_start_time
    state.end_time = wsstate.path_end_time
    state.show_ifnames = wsstate.path_show_ifnames
    if wsstate.path_swap:
        state.source, state.dest = state.dest, state.source
        wsstate.source = state.source
        wsstate.dest = state.dest


def path_run():

    state = st.session_state.pathSessionState
    sqobjs = st.session_state.sqobjs

    if not all(x for x in [state.source, state.dest, state.namespace]):
        return

    pgbar = st.empty()
    summary = st.container()
    summcol, mid, pathcol = summary.columns([3, 1, 10])
    with summary:
        with summcol:
            summ_ph = st.empty()
            legend_ph = st.container()
        with pathcol:
            fw_ph = st.empty()

    pgbar.progress(0)
    pathobj = PathObj(start_time=state.start_time,
                      end_time=state.end_time)
    try:
        df, summ_df = path_get(state, pathobj, forward_dir=True)
        rdf = getattr(pathobj.engine, '_rdf', pd.DataFrame())
        if not rdf.empty:
            state.pathobj = pathobj
            state.path_df = df
    except Exception as e:
        st.error(f'Invalid Input: {str(e)}')
        st.stop()
    pgbar.progress(40)
    # rev_df, _ = path_get(state, forward_dir=False)

    if df.empty:
        pgbar.progress(100)
        st.info(f'No path to trace between {state.source} and {state.dest}')
        st.experimental_set_query_params(**get_path_url_params(state))
        st.stop

    if not df.empty:
        faileddfs = get_failed_data(state.namespace, pgbar,
                                    st.session_state.sqobjs)
        g = build_graphviz_obj(state.show_ifnames, df, faileddfs)
        pgbar.progress(100)
        # if not rev_df.empty:
        #     rev_g = build_graphviz_obj(state, rev_df)

        summ_ph.dataframe(data=summ_df.astype(str))
        with legend_ph:
            st.info('''Color Legend''')
            st.markdown('''
<b style="color:blue">Blue Lines</b> => L2 Hop(non-tunneled)<br>
<b>Black Lines</b> => L3 Hop<br>
<b style="color:purple">Purple lines</b> => Tunneled Hop<br>
<b style="color:red">Red Lines</b> => Hops with Error<br>
''', unsafe_allow_html=True)

        fw_ph.graphviz_chart(g, use_container_width=True)
        # rev_ph.graphviz_chart(rev_g, use_container_width=True)

        for entry in faileddfs.dfs:
            mdf = entry['df']
            table_expander = st.expander(
                f'Failed {entry["name"]} Table', expanded=not mdf.empty)
            with table_expander:
                st.dataframe(mdf)

        table_expander = st.expander('Path Table', expanded=True)
        with table_expander:
            st.dataframe(data=df.style.apply(highlight_erroneous_rows, axis=1),
                         height=600)

    st.experimental_set_query_params(**get_path_url_params(state))


def page_work(state_container):
    '''Main workhorse routine for path'''

    if 'pathSessionState' not in state_container:
        state_container['pathSessionState'] = None

    state = state_container.pathSessionState

    url_params = st.experimental_get_query_params()
    page = url_params.pop('page', '')

    if not state and get_title() in page:
        if url_params and not all(not x for x in url_params.values()):
            url_params.pop('search_text', '')
            for key in url_params:
                val = url_params.get(key, '')
                if isinstance(val, list):
                    val = val[0]
                    url_params[key] = val
                if key in ['show_ifnames']:
                    if val == 'True':
                        url_params[key] = True
                    else:
                        url_params[key] = False

            state = PathSessionState(**url_params)
            state_container.pathSessionState = state
        else:
            state = PathSessionState()
            state_container.pathSessionState = state
    elif not state:
        state = PathSessionState()
        state_container.pathSessionState = state

    path_sidebar(state, state_container.sqobjs)
    path_run()



#!/usr/bin/env python3

import sys
import shlex
import yaml
from os.path import exists


def process_ansible_inventory(filename, namespace='default'):
    """Read an Ansible inventory file and produce device list for suzieq

    :param filename: Ansible inventory filename
    :type filename: string

    :param namespace: The namespace associated with the data from these devices
    :type namespace: string

    :rtype: list:
    :return: List of devices as extracted from the Ansible file
    """

    with open(filename, 'r') as f:
        lines = f.readlines()

    hostsdata = []
    for line in lines:
        host = None
        hostline = ''
        out = shlex.split(line)
        if 'ansible_network_os=eos' in out:
            transport = 'https'
            addnl_info = 'devtype=eos'
        else:
            transport = 'ssh'
            addnl_info = ''
        for elem in out:
            if elem.startswith('ansible_host='):
                k, v = elem.split('=')

                if v == '127.0.0.1':
                    host = v
                    continue
                else:
                    hostline = ('    - url: {}://vagrant@{} {}'
                                .format(transport, v, addnl_info))
            if elem.startswith('ansible_port='):
                k, v = elem.split('=')
                if host:
                    hostline = (
                        '    - url: ssh://vagrant@{}:{}'.format(host, v))

            if elem.startswith('ansible_ssh_private_key_file'):
                k, v = elem.split('=')
                if not exists(v.strip()):
                    continue
                if hostline:
                    hostline += ' keyfile={}'.format(v.strip())
                else:
                    addnl_info += ' keyfile={}'.format(v.strip())

        hostsdata.append(hostline)

    hostsdata.insert(0, '- namespace: {}'.format(namespace))
    hostsdata.insert(1, '  hosts:')

    return hostsdata


def convert_ansible_inventory(filename: str, namespace: str = 'default'):
    """Converts the output of ansible-inventory command for processing.

    Ansible pulls together the inventory information from multiple files. The
    information relevant to sq-poller maybe present in different files as
    different vars. ansible-inventory command luckily handles this for us. This
    function takes the JSON output of that command and gathers the data needed
    to start polling.

    Parameters
    ----------
    :pararm filename: file containing the output of ansible-inventory output
    :type filename" string

    :param namespace: namespace associated with this inventory
    :type namespace: string

    Returns
    -------
    :rtype: list
    :return: List of devices and the related info to connect
    """

    try:
        with open(filename, 'r') as f:
            inventory = yaml.safe_load(f)
    except Exception as error:
        print(f':ERROR: Unable to process Ansible inventory: {str(error)}')
        sys.exit(1)

    if '_meta' not in inventory or "hostvars" not in inventory['_meta']:
        if isinstance(inventory, list) and 'namespace' in inventory[0]:
            print("ERROR: Invalid Ansible inventory, found Suzieq inventory, "
                  "use -D option")
        else:
            print("ERROR: Invalid Ansible inventory, "
                  "missing keys: _meta and / or hostvars\n"
                  "\tUse 'ansible-inventory --list' to create the correct file")
        sys.exit(1)

    in_hosts = inventory['_meta']['hostvars']
    out_hosts = []
    for host in in_hosts:
        entry = in_hosts[host]
        addnl_info = ''

        addnl_info += f'username={entry["ansible_user"]} '
        if 'ansible_password' in entry:
            addnl_info += f' password={entry["ansible_password"]} '
        if entry.get('ansible_network_os', '') == 'eos':
            transport = 'https://'
            addnl_info += 'devtype=eos'
            port = 443
        else:
            transport = 'ssh://'
            port = entry.get("ansible_port", 22)

        url = (f'{transport}{entry["ansible_host"]}'
               f':{port}')
        keyfile = entry.get('ansible_ssh_private_key_file', '')
        if keyfile:
            if exists(keyfile.strip()):
                hostline = f'    - url: {url} keyfile={keyfile} {addnl_info}'
            else:
                hostline = f'    - url: {url} {addnl_info}'
        else:
            hostline = f'    - url: {url} {addnl_info}'

        out_hosts.append(hostline)

    out_hosts.insert(0, '- namespace: {}'.format(namespace))
    out_hosts.insert(1, '  hosts:')

    return out_hosts


if __name__ == '__main__':
    if len(sys.argv) < 4:
        print('Usage: genhosts <Ansible inventory file> <output file> <DC name>')
        sys.exit(1)

    hostsdata = convert_ansible_inventory(sys.argv[1], sys.argv[3])

    out = '\n'.join(hostsdata)

    with open(sys.argv[2], 'w') as f:
        f.write(out)



from suzieq.sqobjects.basicobj import SqObject


class FsObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='fs', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'mountPoint', 'usedPercent', 'query_str']



from pygments.token import Token

from nubia import context
from nubia import statusbar
from pandas import DataFrame

from suzieq.version import SUZIEQ_VERSION


class NubiaSuzieqStatusBar(statusbar.StatusBar):
    def __init__(self, ctx):
        self._last_status = None
        self.ctx = ctx

    def get_rprompt_tokens(self):
        if not isinstance(self._last_status, DataFrame) and self._last_status:
            return [(Token.RPrompt, "Error: {}".format(self._last_status))]
        return []

    def set_last_command_status(self, status):
        self._last_status = status

    def get_tokens(self):
        spacer = (Token.Spacer, "  ")

        if context.get_context().pager:
            is_pager = (Token.Warn, "ON")
        else:
            is_pager = (Token.Info, "OFF")

        return [
            (Token.Toolbar, "Suzieq"),
            spacer,
            (Token.Toolbar, "Version "),
            spacer,
            (Token.Info, SUZIEQ_VERSION),
            spacer,
            (Token.Toolbar, "Pager "),
            spacer,
            is_pager,
            spacer,
            (Token.Toolbar, "Namespace "),
            spacer,
            (Token.Info, ", ".join(self.ctx.namespace)),
            spacer,
            (Token.Toolbar, "Hostname "),
            spacer,
            (Token.Info, ", ".join(self.ctx.hostname)),
            spacer,
            (Token.Toolbar, "StartTime "),
            spacer,
            (Token.Info, self.ctx.start_time),
            spacer,
            (Token.Toolbar, "EndTime "),
            spacer,
            (Token.Info, self.ctx.end_time),
            spacer,
            (Token.Toolbar, "Engine "),
            spacer,
            (Token.Info, self.ctx.engine),
            spacer,
            (Token.Toolbar, "Query Time "),
            spacer,
            (Token.Info, self.ctx.exec_time),
        ]



import os
from importlib.util import find_spec
from importlib import import_module
import inspect


def get_engine_object(table, baseobj):

    spec = find_spec('suzieq.engines.pandas')
    for file in spec.loader.contents():
        if (os.path.isfile(f'{os.path.dirname(spec.loader.path)}/{file}') and
                not file.startswith('_')):
            modname = file.split('.')[0]
            mod = import_module(f'suzieq.engines.pandas.{modname}')
            for mbr in inspect.getmembers(mod):
                if inspect.isclass(mbr[1]) and mbr[0] != 'SqPandasEngine':
                    fn = getattr(mbr[1], 'table_name', '')
                    if fn and fn() == table:
                        return mbr[1](baseobj)

    return None


__all__ = [get_engine_object]



from pandas import DataFrame
from suzieq.sqobjects.basicobj import SqObject


class PathObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='path', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'vrf', 'source', 'dest']

    def summarize(self, namespace=[], hostname=[], source='',
                  dest='', vrf='', query_str='') -> DataFrame():
        """Path summarize, different because of the params

        :param namespace: List{str], can really only be a single namespace
        :param hostname: List[str], ignored for now
        :param source: str, Source IP address,
        :param dest: str, Dest IP address,
        :param vrf: str, VRF within which to run the path
        :param query_str: str, Pandas query string to further refine result
        :returns: Pandas dataframe, summarizing the result
        """
        if self.columns != ["default"]:
            self.summarize_df = DataFrame(
                {'error': ['ERROR: You cannot specify columns with summarize']})
            return self.summarize_df
        if not self._table:
            raise NotImplementedError

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        return self.engine.summarize(namespace=namespace, source=source,
                                     dest=dest, vrf=vrf, query_str=query_str)



import pandas as pd
import numpy as np
from suzieq.utils import SchemaForTable, humanize_timestamp
from suzieq.engines.base_engine import SqEngineObj
from suzieq.db import get_sqdb_engine
from suzieq.exceptions import DBReadError, UserQueryError
import dateparser
from datetime import datetime
from pandas.core.groupby import DataFrameGroupBy


class SqPandasEngine(SqEngineObj):
    def __init__(self, baseobj):
        self.ctxt = baseobj.ctxt
        self.iobj = baseobj
        self.summary_row_order = []
        self._summarize_on_add_field = []
        self._summarize_on_add_with_query = []
        self._summarize_on_add_list_or_count = []
        self._summarize_on_add_stat = []
        self._summarize_on_perdevice_stat = []
        self._dbeng = get_sqdb_engine(baseobj.ctxt.cfg, baseobj.table, '',
                                      None)

    @property
    def schemas(self):
        return self.ctxt.schemas

    @property
    def cfg(self):
        return self.iobj._cfg

    @property
    def table(self):
        return self.iobj._table

    def _get_ipvers(self, value: str) -> int:
        """Return the IP version in use"""

        if ':' in value:
            ipvers = 6
        elif '.' in value:
            ipvers = 4
        else:
            ipvers = ''

        return ipvers

    def _handle_user_query_str(self, df: pd.DataFrame,
                               query_str: str) -> pd.DataFrame:
        """Handle user query, trapping errors and returning exception

        Args:
            df (pd.DataFrame): The dataframe to run the query on
            query_str (str): pandas query string

        Raises:
            UserQueryError: Exception if pandas query aborts with errmsg

        Returns:
            pd.DataFrame: dataframe post query
        """
        if query_str:
            if query_str.startswith('"') and query_str.endswith('"'):
                query_str = query_str[1:-1]

            try:
                df = df.query(query_str).reset_index(drop=True)
            except Exception as ex:
                raise UserQueryError(ex)

        return df

    def get_valid_df(self, table: str, **kwargs) -> pd.DataFrame:
        """The heart of the engine: retrieving the data from the backing store

        Args:
            table (str): Name of the table to retrieve the data for

        Returns:
            pd.DataFrame: The data as a pandas dataframe
        """
        if not self.ctxt.engine:
            print("Specify an analysis engine using set engine command")
            return pd.DataFrame(columns=["namespace", "hostname"])

        sch = SchemaForTable(table, schema=self.schemas)
        phy_table = sch.get_phy_table_for_table()

        columns = kwargs.pop('columns', ['default'])
        addnl_fields = kwargs.pop('addnl_fields', [])
        view = kwargs.pop('view', self.iobj.view)
        active_only = kwargs.pop('active_only', True)
        hostname = kwargs.get('hostname', [])

        fields = sch.get_display_fields(columns)
        key_fields = sch.key_fields()
        drop_cols = []

        if columns == ['*']:
            drop_cols.append('sqvers')

        aug_fields = sch.get_augmented_fields()

        if 'timestamp' not in fields:
            fields.append('timestamp')

        if 'active' not in fields+addnl_fields:
            addnl_fields.append('active')
            drop_cols.append('active')

        # Order matters. Don't put this before the missing key fields insert
        for f in aug_fields:
            dep_fields = sch.get_parent_fields(f)
            addnl_fields += dep_fields

        for fld in key_fields:
            if fld not in fields+addnl_fields:
                addnl_fields.insert(0, fld)
                drop_cols.append(fld)

        for f in addnl_fields:
            if f not in fields:
                # timestamp is always the last field
                fields.insert(-1, f)

        if self.iobj.start_time:
            try:
                start_time = int(dateparser.parse(
                    self.iobj.start_time.replace('last night', 'yesterday'))
                    .timestamp()*1000)
            except Exception as e:
                print(f"ERROR: invalid time {self.iobj.start_time}: {e}")
                return pd.DataFrame()
        else:
            start_time = ''

        if self.iobj.start_time and not start_time:
            # Something went wrong with our parsing
            print(f"ERROR: unable to parse {self.iobj.start_time}")
            return pd.DataFrame()

        if self.iobj.end_time:
            try:
                end_time = int(dateparser.parse(
                    self.iobj.end_time.replace('last night', 'yesterday'))
                    .timestamp()*1000)
            except Exception as e:
                print(f"ERROR: invalid time {self.iobj.end_time}: {e}")
                return pd.DataFrame()
        else:
            end_time = ''

        if self.iobj.end_time and not end_time:
            # Something went wrong with our parsing
            print(f"ERROR: Unable to parse {self.iobj.end_time}")
            return pd.DataFrame()

        table_df = self._dbeng.read(
            phy_table,
            'pandas',
            start_time=start_time,
            end_time=end_time,
            columns=fields,
            view=view,
            key_fields=key_fields,
            **kwargs
        )

        if not table_df.empty:
            # hostname may not have been filtered if using regex
            if hostname:
                hdf_list = []
                for hn in hostname:
                    df1 = table_df.query(f"hostname.str.match('{hn}')")
                    if not df1.empty:
                        hdf_list.append(df1)

                if hdf_list:
                    table_df = pd.concat(hdf_list)

            if view == "all" or not active_only:
                table_df.drop(columns=drop_cols, inplace=True)
            else:
                table_df = table_df.query('active') \
                                   .drop(columns=drop_cols)
            if 'timestamp' in table_df.columns and not table_df.empty:
                table_df['timestamp'] = humanize_timestamp(
                    table_df.timestamp, self.cfg.get('analyzer', {})
                    .get('timezone', None))

        return table_df

    def get(self, **kwargs) -> pd.DataFrame:
        """The default get method for all tables

        Use this for a table if nothing special is desired. No table uses
        this routine today.

        Raises:
            NotImplementedError: If no table has been defined

        Returns:
            pd.DataFrame: pandas dataframe of the object
        """
        if not self.iobj.table:
            raise NotImplementedError

        user_query = kwargs.pop('query_str', '')

        df = self.get_valid_df(self.iobj.table, **kwargs)

        df = self._handle_user_query_str(df, user_query)

        return df

    def get_table_info(self, table: str, **kwargs) -> dict:
        """Returns information about the data available for a table

        Used by table show command exclusively.
        Args:
            table (str): Name of the table about which info is desired

        Returns:
            dict: The desired data as a dictionary
        """
        # You can't use view from user because we need to see all the data
        # to compute data required.
        kwargs.pop('view', None)

        all_time_df = self.get_valid_df(table, view='all', **kwargs)
        times = all_time_df['timestamp'].unique()
        ret = {'firstTime': all_time_df.timestamp.min(),
               'latestTime': all_time_df.timestamp.max(),
               'intervals': len(times),
               'allRows': len(all_time_df),
               'namespaces': self._unique_or_zero(all_time_df, 'namespace'),
               'deviceCnt': self._unique_or_zero(all_time_df, 'hostname')}

        return ret

    def _unique_or_zero(self, df: pd.DataFrame, col: str) -> int:
        """Returns the unique count of a column in a dataframe or 0

        Args:
            df (pd.DataFrame): The dataframe to use
            col (str): The column name to use

        Returns:
            int: Count of unique values
        """
        if col in df.columns:
            return df[col].nunique()
        else:
            return 0

    def summarize(self, **kwargs):
        """Summarize the info about this resource/service.

        There is a pattern of how to do these
        use self._init_summarize():
           - creates self.summary_df, which is the initial pandas dataframe
             based on the table
           - creates self.nsgrp of data grouped by namespace
           - self.ns is the dict to add data to which will be turned into a
             dataframe and then returned

        if you want to simply take a field and run a pandas functon, then use
          self._add_field_to_summary

        at the end of te summarize
        return pd.DataFrame(self.ns).convert_dtypes()

        If you don't override this, then you get a default summary of all columns
        """
        self._init_summarize(self.iobj._table, **kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._gen_summarize_data()
        self._post_summarize()
        return self.ns_df.convert_dtypes()

    def unique(self, **kwargs) -> pd.DataFrame:
        """Return the unique elements as per user specification

        Raises:
            ValueError: If len(columns) != 1

        Returns:
            pd.DataFrame: Pandas dataframe of unique values for given column
        """
        count = kwargs.pop("count", 0)
        query_str = kwargs.get('query_str', '')

        columns = kwargs.pop("columns", None)

        if query_str:
            getcols = ['*']
        else:
            getcols = columns

        column = columns[0]

        df = self.get(columns=getcols, **kwargs)
        if df.empty:
            return df

        # check if column we're looking at is a list, and if so explode it
        if df.apply(lambda x: isinstance(x[column], np.ndarray), axis=1).all():
            df = df.explode(column).dropna(how='any')

        if not count:
            return (pd.DataFrame({f'{column}': df[column].unique()}))
        else:
            r = df[column].value_counts()
            return (pd.DataFrame({column: r})
                    .reset_index()
                    .rename(columns={column: 'numRows',
                                     'index': column})
                    .sort_values(column))

    def analyze(self, **kwargs):
        raise NotImplementedError

    def aver(self, **kwargs):
        raise NotImplementedError

    def top(self, **kwargs):
        """Default implementation of top.
        The basic fields this assumes are present include the "what" keyword
        which contains the name of the field we're getting the transitions on,
        the "n" field which tells the count of the top entries you're
        looking for, and the reverse field which tells whether you're looking
        for the largest (default, and so reverse is False) or the smallest(
        reverse is True). This invokes the default object's get routine. It
        is upto the caller to ensure that the desired column is in the output.
        """
        what = kwargs.pop("what", None)
        reverse = kwargs.pop("reverse", False)
        sqTopCount = kwargs.pop("count", 5)

        if not what:
            return pd.DataFrame()

        df = self.get(addnl_fields=self.iobj._addnl_fields, **kwargs)
        if df.empty or ('error' in df.columns):
            return df

        if reverse:
            return df.nsmallest(sqTopCount, columns=what, keep="all") \
                     .head(sqTopCount)
        else:
            return df.nlargest(sqTopCount, columns=what, keep="all") \
                     .head(sqTopCount)

    def _init_summarize(self, table: str, **kwargs) -> None:
        """Initialize the data structures for use with generating summary

        Args:
            table (str): Name of table to generate summary for
        """
        kwargs.pop('columns', None)
        columns = ['*']

        df = self.get(columns=columns, **kwargs)
        if 'error' in df.columns:
            self.summary_df = df
            return

        self.summary_df = df
        if df.empty:
            return

        self.ns = {i: {} for i in df['namespace'].unique()}
        self.nsgrp = df.groupby(by=["namespace"], observed=True)

    def _gen_summarize_data(self):
        """Generate the data required for summary"""

        if not self._summarize_on_add_field:
            # Add the only field we truly know to add
            self._summarize_on_add_field = [
                ('deviceCnt', 'hostname', 'nunique'),
            ]
        for field_name, col, function in self._summarize_on_add_field:
            if col != 'namespace' and col != 'timestamp':
                self._add_field_to_summary(col, function, field_name)
                self.summary_row_order.append(field_name)

        for flds in self._summarize_on_add_with_query:
            field_name = flds[0]
            query_str = flds[1]
            field = flds[2]
            if len(flds) == 4:
                func = flds[3]
            else:
                func = 'count'
            fld_df = self.summary_df.query(query_str)
            if not fld_df.empty:
                fld_per_ns = fld_df.groupby(by=['namespace'],
                                            observed=True)[field] \
                    .agg(func)
                for i in self.ns.keys():
                    self.ns[i].update({field_name: fld_per_ns.get(i, 0)})
            else:
                for i in self.ns.keys():
                    self.ns[i].update({field_name: 0})

            self.summary_row_order.append(field_name)

        for field_name, field in self._summarize_on_add_list_or_count:
            self._add_list_or_count_to_summary(field, field_name)
            self.summary_row_order.append(field_name)

        for field_name, query_str, field in self._summarize_on_add_stat:
            if query_str:
                statfld = self.summary_df.query(query_str) \
                                         .groupby(by=['namespace'],
                                                  observed=True)[field]
            else:
                statfld = self.summary_df.groupby(
                    by=['namespace'], observed=True)[field]

            self._add_stats_to_summary(statfld, field_name)
            self.summary_row_order.append(field_name)

        for field_name, query_str, field, func in \
                self._summarize_on_perdevice_stat:
            if query_str:
                statfld = self.summary_df \
                    .query(query_str) \
                    .groupby(by=['namespace', 'hostname'],
                             observed=True)[field].agg(func)
            else:
                statfld = self.summary_df \
                    .groupby(by=['namespace', 'hostname'],
                             observed=True)[field].agg(func)

            self._add_stats_to_summary(statfld, field_name, filter_by_ns=True)
            self.summary_row_order.append(field_name)

    def _post_summarize(self, check_empty_col: str = 'deviceCnt') -> None:
        """Once summary data has been generated, finalize summary dataframe

        Args:
            check_empty_col (str, optional): column name to check to remove
                                             namespace that's empty. 
                                             Defaults to 'deviceCnt'.
        """
        # this is needed in the case that there is a namespace that has no
        # data for this command

        if not check_empty_col:
            check_empty_col = self._check_empty_col

        delete_keys = []
        for ns in self.ns:
            if self.ns[ns][check_empty_col] == 0:
                delete_keys.append(ns)
        for ns in delete_keys:
            del(self.ns[ns])

        ns_df = pd.DataFrame(self.ns)
        if len(self.summary_row_order) > 0:
            ns_df = ns_df.reindex(self.summary_row_order, axis=0)
        self.ns_df = ns_df

    def _add_constant_to_summary(self, field: str, value):
        """And a constant value to specified field name in summary"""
        if field:
            {self.ns[i].update({field: value}) for i in self.ns.keys()}

    def _add_field_to_summary(self, field: str, method: str = 'nunique',
                              field_name: str = None):
        """Add a summary field to summary dataframe based on method on field

        This assumes that self.summary_df has been populated.
        Args:
            field (str): Column name to add in the summary dataframe
            method (str, optional): pandas method name. Defaults to 'nunique'.
            field_name (str, optional): column name to compute on. Defaults to None.
        """
        if not field_name:
            field_name = field
        field_per_ns = getattr(self.nsgrp[field], method)()
        {self.ns[i].update({field_name: field_per_ns.get(i, 0)})
         for i in self.ns.keys()}

    def _add_list_or_count_to_summary(self, field: str,
                                      field_name: str = None) -> None:
        """if there are less than 3 unique things, add as a list, otherwise return the count"""
        if not field_name:
            field_name = field

        count_per_ns = self.nsgrp[field].nunique()

        for n in self.ns.keys():
            if 3 >= count_per_ns[n] > 0:
                # can't do a value_counts on all groups, incase one of the groups other groups doesn't have data
                unique_for_ns = self.nsgrp.get_group(n)[field].value_counts()
                value = unique_for_ns.to_dict()
                # Filter numm entries if category because of how pandas behaves here
                if self.nsgrp[field].dtype[n].name == 'category':
                    value = dict(filter(lambda x: x[1] != 0, value.items()))

            else:
                value = count_per_ns[n]

            self.ns[n].update({field_name: value})

    def _add_stats_to_summary(self, groupedby: DataFrameGroupBy,
                              fieldname: str,
                              filter_by_ns: bool = False) -> None:
        """Takes grouped stats and adds min, max, and median to stats"""

        {self.ns[i].update({fieldname: []}) for i in self.ns.keys()}
        if filter_by_ns:
            {self.ns[i][fieldname].append(groupedby[i].min())
             if i in groupedby else self.ns[i][fieldname].append(0)
             for i in self.ns.keys()}
            {self.ns[i][fieldname].append(groupedby[i].max())
             if i in groupedby else self.ns[i][fieldname].append(0)
             for i in self.ns.keys()}
            {self.ns[i][fieldname].append(
                groupedby[i].median(numeric_only=False))
             if i in groupedby else self.ns[i][fieldname].append(0)
             for i in self.ns.keys()}
        else:
            min_field = groupedby.min()
            max_field = groupedby.max()
            med_field = groupedby.median(numeric_only=False)

            {self.ns[i][fieldname].append(min_field[i])
             if i in min_field else self.ns[i][fieldname].append(0)
             for i in self.ns.keys()}
            {self.ns[i][fieldname].append(max_field[i])
             if i in max_field else self.ns[i][fieldname].append(0)
             for i in self.ns.keys()}
            {self.ns[i][fieldname].append(med_field[i])
             if i in med_field else self.ns[i][fieldname].append(0)
             for i in self.ns.keys()}



from importlib.util import find_spec
from importlib import import_module
import inspect


name = "sqobjects"


def get_tables():
    fspec = find_spec('suzieq.sqobjects')
    tables = [x.split('.')[0] for x in fspec.loader.contents()
              if not x.startswith('_') and x != "basicobj.py"]
    tables.extend(['tables'])
    return tables


def get_sqobject(table_name: str):
    '''Return the object to operate on the table requested

    :param table_name: str, name of table for which object is requested
    :returns:
    :rtype: The child class of SqObject that corresponds to the table
            or raises ModuleNotFoundError if the table is not found

    '''

    try:
        mod = import_module(f'suzieq.sqobjects.{table_name}')
        for mbr in inspect.getmembers(mod):
            if inspect.isclass(mbr[1]):
                if getattr(mbr[1], '__bases__')[0].__name__ == 'SqObject':
                    return mbr[1]
    except ModuleNotFoundError:
        return None

    raise ModuleNotFoundError(f'{table_name} not found')


__all__ = get_tables() + ['get_sqobject']

sqobjs_all = __all__



from itertools import zip_longest
from typing import Collection

import pandas as pd
import streamlit as st


def get_title():
    # suzieq_gui.py has hardcoded this name.
    return '_Path_Debug_'


def path_debug_sidebar(state):
    '''Draw the sidebar'''

    st.sidebar.markdown(
        """Displays information from the various tables used to build path""")


def handle_edge_url(url_params: dict, pathSession):
    '''Display tables associated with a link'''

    namespace = url_params.get('namespace', [""])[0]
    hostname = url_params.get('hostname', [""])[0]
    nhip = url_params.get('nhip', [""])[0]
    ipLookup = url_params.get('ipLookup', [""])[0]
    vtepLookup = url_params.get('vtepLookup', [""])[0]
    vrf = url_params.get('vrf', [""])[0]
    ifhost = url_params.get('ifhost', [""])[0]
    macaddr = url_params.get('macaddr', [""])[0]
    oif = url_params.get('oif', [""])[0]

    if not hostname:
        st.error('No hostname found to display information for')
        st.stop()

    st.header(f'Debug Tables for Path from {pathSession.source} to '
              f'{pathSession.dest}')
    hoptype = 'Bridged' if macaddr else 'Routed'
    st.subheader(f'{hoptype} hop between {hostname} and {ifhost}')

    pathobj = getattr(pathSession, 'pathobj', None)
    engobj = pathobj.engine

    if ipLookup:
        if not vtepLookup or (ipLookup != vtepLookup):
            st.info(f'Route Lookup on {hostname}')
            st.dataframe(data=engobj._rdf.query(
                f'hostname=="{hostname}" and vrf=="{vrf}"'))

        if vtepLookup:
            st.info(f'Underlay Lookup on {hostname} for {vtepLookup}')
            vtepdf = engobj._underlay_dfs.get(vtepLookup,
                                              pd.DataFrame())
            if not vtepdf.empty:
                st.dataframe(data=vtepdf.query(
                    f'hostname=="{hostname}" and vrf=="default"'))
        if nhip:
            st.info(
                f'ARP/ND Table on {hostname} for nexthop {nhip}, oif={oif}')
            arpdf = engobj._arpnd_df.query(f'hostname=="{hostname}" and '
                                           f'ipAddress=="{nhip}" and '
                                           f'oif=="{oif}"')
            st.dataframe(data=arpdf)

            if ':' in nhip:
                dropcol = ['ipAddressList']
            else:
                dropcol = ['ip6AddressList']
            if not arpdf.empty:
                nhmac = arpdf.macaddr.iloc[0]
                if nhmac:
                    if_df = engobj._if_df.query(f'macaddr=="{nhmac}" and '
                                                f'hostname=="{ifhost}"') \
                                         .drop(columns=dropcol)
                    label = (f'matching nexthop {nhip}, macaddr {nhmac} on '
                             f'host {ifhost}')
                else:
                    label = f'matching nexthop {nhip} on host {ifhost}'
                    if_df = engobj._if_df.query(f'hostname=="{ifhost}"') \
                                         .drop(columns=dropcol)
            else:
                if_df = engobj._if_df.query(f'hostname=="{ifhost}"')\
                    .drop(columns=dropcol)
                label = f'matching host {ifhost}'
            if nhip != '169.254.0.1':
                st.info(f'Interfaces {label}')
                s = if_df.ipAddressList.str \
                                       .startswith(f'{nhip}/') \
                                       .dropna()
                s = s.loc[s == True]
                st.dataframe(data=engobj._if_df.iloc[s.loc[s == True].index])
            else:
                st.info(f'Interfaces {label}')
                st.dataframe(data=if_df)
    if macaddr:
        with st.expander(f'MAC Table for {hostname}, MAC addr {macaddr}',
                         expanded=True):
            st.dataframe(data=pathobj.engine._macsobj.get(
                namespace=namespace, hostname=hostname, macaddr=macaddr))


def handle_hop_url(url_params, pathSession):
    '''Handle table display associated with hop'''

    namespace = url_params.get('namespace', [""])[0]
    hostname = url_params.get('hostname', [""])[0]

    if not hostname:
        st.error('No hostname found to display information for')
        st.stop()

    st.header(f'Debug Tables for Path from {pathSession.source} to '
              f'{pathSession.dest}')

    pathobj = getattr(pathSession, 'pathobj', None)
    df = getattr(pathSession, 'path_df', None)
    engobj = pathobj.engine

    if df.empty:
        st.warning('Empty path dataframe')
        st.stop()

    host_dfg = df.query(f'hostname == "{hostname}"') \
                 .groupby(by=['hopCount'])

    df2 = host_dfg.agg({'vrf': ['unique'], 'ipLookup': ['unique'],
                        'nexthopIp': ['unique'], 'oif': ['unique'],
                        'macLookup': ['unique'],
                        'vtepLookup': ['unique']}).reset_index()
    df2.columns = ['hopCount', 'vrf', 'ipLookup', 'nexthopIp', 'oif',
                   'macaddr', 'vtepLookup']
    df2 = df2.explode('hopCount').explode('vrf').explode('ipLookup') \
                                                .explode('macaddr') \
                                                .explode('vtepLookup')
    df2.drop_duplicates(subset=['vrf', 'ipLookup'], inplace=True)

    for row in df2.itertuples():
        with st.expander(f'Lookups on {hostname}, for hopcount: '
                         f'{row.hopCount}', expanded=True):
            if row.macaddr:
                st.info(f'MAC Table on {hostname}, MAC addr {row.macaddr}')
                st.dataframe(data=engobj._macsobj.get(namespace=namespace,
                                                      hostname=hostname,
                                                      macaddr=row.macaddr))
                continue

            if (row.ipLookup != row.vtepLookup):
                st.info(f'Route Lookup on {hostname}')
                st.dataframe(data=engobj._rdf.query(
                    f'hostname=="{hostname}" and vrf=="{row.vrf}"'))

            if row.vtepLookup:
                st.info(f'Underlay Lookup on {hostname} for {row.vtepLookup}')
                vtepdf = engobj._underlay_dfs.get(row.vtepLookup,
                                                  pd.DataFrame())
                if not vtepdf.empty:
                    st.dataframe(data=vtepdf.query(
                        f'hostname=="{hostname}" and vrf=="default"'))

            oifs = row.oif.tolist()
            nhops = row.nexthopIp.tolist()
            prev_nhop = ''
            for oif, nhop in zip_longest(oifs, nhops):
                blank1, arpcol = st.columns([1, 40])
                blank2, ifcol = st.columns([2, 40])
                # this logic because I don't know what fn to use with agg above
                # to not remove non-unique nhop.
                if not nhop and prev_nhop:
                    nhop = prev_nhop
                else:
                    prev_nhop = nhop
                arpdf = engobj._arpnd_df.query(f'hostname=="{hostname}" and '
                                               f'ipAddress=="{nhop}" and '
                                               f'oif=="{oif}"')
                with arpcol:
                    st.info(f'ARP/ND Lookup on {hostname} for {nhop}')
                    st.dataframe(data=arpdf, height=100)

                if not arpdf.empty:
                    if ':' in nhop:
                        dropcol = ['ipAddressList']
                    else:
                        dropcol = ['ip6AddressList']
                    if nhop == '169.254.0.1':
                        macaddr = arpdf.macaddr.iloc[0]
                        if_df = engobj._if_df.query(f'macaddr=="{macaddr}"') \
                                             .drop(columns=dropcol)
                        label = f'matching nexthop {nhop}, macaddr {macaddr}'
                    else:
                        if_df = engobj._if_df.drop(columns=dropcol)
                        label = f'matching nexthop {nhop}'
                else:
                    label = f'matching nexthop {nhop}'
                    if_df = engobj._if_df
                if ':' in nhop:
                    s = if_df.ip6AddressList \
                             .explode() \
                             .str.startswith(f'{nhop}/').dropna()
                    s = s.loc[s == True]
                    if_df = if_df.iloc[s.loc[s == True].index]
                elif nhop != '169.254.0.1':
                    s = if_df.ipAddressList \
                             .explode() \
                             .str.startswith(f'{nhop}/').dropna()
                    s = s.loc[s == True]
                    if_df = if_df.iloc[s.loc[s == True].index]
                with ifcol:
                    st.info(f'Interfaces {label}')
                    st.dataframe(data=if_df, height=600)
        st.markdown("<hr>", unsafe_allow_html=True)


def page_work(state_container):
    '''Main page workhorse'''

    pathSession = state_container.pathSessionState

    if pathSession:
        pathobj = getattr(pathSession, 'pathobj', None)
    else:
        st.error('No saved path session found.')
        st.stop()

    if not pathobj:
        st.error('No saved path object found.')
        st.stop()

    path_debug_sidebar(pathSession)

    url_params = st.experimental_get_query_params()
    if url_params.get('lookupType', 'hop') == ['hop']:
        handle_hop_url(url_params, pathSession)
    else:
        handle_edge_url(url_params, pathSession)



from os.path import dirname, basename, isfile, join
import glob

name = "pages"

modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [basename(f)[:-3] for f in modules if isfile(f)
           and not f.endswith('__init__.py')]

sq_gui_pages_all = __all__



from typing import List

import pandas as pd
from suzieq.sqobjects.basicobj import SqObject


class TopologyObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='topology', **kwargs)
        self._sort_fields = ["namespace", "hostname", "ifname"]
        self._cat_fields = []
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'polled', 'ifname', 'via', 'peerHostname',
                                'query_str']
        self._valid_summarize_args = ['namespace', 'hostname', 'via',
                                      'query_str']
        self._valid_arg_vals = {
            'polled': ['True', 'False', 'true', 'false', '']
        }

    def summarize(self, namespace: List[str] = [], hostname: List[str] = [],
                  via: List[str] = [], query_str='') -> pd.DataFrame:
        if self.columns != ["default"]:
            self.summarize_df = pd.DataFrame(
                {'error': ['ERROR: You cannot specify columns with summarize']})
            return self.summarize_df
        if not self._table:
            raise NotImplementedError

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        return self.engine.summarize(namespace=namespace, hostname=hostname,
                                     via=via, query_str=query_str)



import time

from nubia import command

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.mlag import MlagObj


@command('mlag', help="Act on mlag data")
class MlagCmd(SqCommand):
    """Multichassis LAG information (includes variants such as NXOX vPC)"""

    def __init__(self, engine: str = '', hostname: str = '',
                 start_time: str = '', end_time: str = '',
                 view: str = '', namespace: str = '',
                 query_str: str = ' ',
                 format: str = "", columns: str = 'default') -> None:
        super().__init__(engine=engine, hostname=hostname,
                         start_time=start_time, end_time=end_time,
                         view=view, namespace=namespace,
                         format=format, query_str=query_str, columns=columns,
                         sqobj=MlagObj)

    @command('show')
    def show(self):
        """Show MLAG info
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ['default']:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname,
                                columns=self.columns, query_str=self.query_str,
                                namespace=self.namespace)
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if not df.empty and 'state' in df.columns:
            return self._gen_output(df.query('state != "disabled"'))
        else:
            return self._gen_output(df)



import subprocess
from importlib.util import find_spec
import os


def gui_main():

    spec = find_spec('suzieq.gui')
    if spec:
        thisprog = f'{os.path.dirname(spec.loader.path)}/suzieq-gui.py'

    _ = subprocess.check_output(f'streamlit run {thisprog}'.split())


if __name__ == '__main__':
    gui_main()



from suzieq.sqobjects.address import AddressObj

import pandas as pd
import numpy as np

from .engineobj import SqPandasEngine
from suzieq.utils import build_query_str, SchemaForTable, humanize_timestamp


class BgpObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'bgp'

    def get(self, **kwargs):
        """Replacing the original interface name in returned result"""

        addnl_fields = kwargs.pop('addnl_fields', [])
        columns = kwargs.get('columns', ['default'])
        vrf = kwargs.pop('vrf', None)
        peer = kwargs.pop('peer', None)
        hostname = kwargs.pop('hostname', None)
        user_query = kwargs.pop('query_str', None)

        drop_cols = ['origPeer', 'peerHost']
        addnl_fields.extend(['origPeer'])
        sch = SchemaForTable(self.iobj.table, self.schemas)
        fields = sch.get_display_fields(columns)

        for col in ['peerIP', 'updateSource', 'state', 'namespace', 'vrf',
                    'peer', 'hostname']:
            if col not in fields:
                addnl_fields.append(col)
                drop_cols.append(col)

        try:
            df = super().get(addnl_fields=addnl_fields, **kwargs)
        except KeyError as ex:
            if ('afi' in str(ex)) or ('safi' in str(ex)):
                df = pd.DataFrame(
                    {'error': [f'ERROR: Migrate BGP data first using sq-coalescer']})
                return df

        if df.empty:
            return df

        if 'afiSafi' in columns or (columns == ['*']):
            df['afiSafi'] = df['afi'] + ' ' + df['safi']
        query_str = build_query_str([], sch, vrf=vrf, peer=peer,
                                    hostname=hostname)
        if 'peer' in df.columns:
            df['peer'] = np.where(df['origPeer'] != "",
                                  df['origPeer'], df['peer'])

        # Convert old data into new 2.0 data format
        if 'peerHostname' in df.columns:
            mdf = self._get_peer_matched_df(df)
            drop_cols = [x for x in drop_cols if x in mdf.columns]
            drop_cols.extend(list(mdf.filter(regex='_y')))
        else:
            mdf = df

        mdf = self._handle_user_query_str(mdf, user_query)

        if query_str:
            return mdf.query(query_str).drop(columns=drop_cols,
                                             errors='ignore')
        else:
            return mdf.drop(columns=drop_cols, errors='ignore')

    def summarize(self, **kwargs) -> pd.DataFrame:
        """Summarize key information about BGP"""

        self._init_summarize(self.iobj._table, **kwargs)
        if self.summary_df.empty or ('error' in self.summary_df.columns):
            return self.summary_df

        self.summary_df['afiSafi'] = (
            self.summary_df['afi'] + ' ' + self.summary_df['safi'])

        afi_safi_count = self.summary_df.groupby(by=['namespace'])['afiSafi'] \
                                        .nunique()

        self.summary_df = self.summary_df \
                              .set_index(['namespace', 'hostname', 'vrf',
                                          'peer']) \
                              .query('~index.duplicated(keep="last")') \
                              .reset_index()
        self.ns = {i: {} for i in self.summary_df['namespace'].unique()}
        self.nsgrp = self.summary_df.groupby(by=["namespace"],
                                             observed=True)

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('totalPeerCnt', 'peer', 'count'),
            ('uniqueAsnCnt', 'asn', 'nunique'),
            ('uniqueVrfsCnt', 'vrf', 'nunique')
        ]

        self._summarize_on_add_with_query = [
            ('failedPeerCnt', 'state == "NotEstd"', 'peer'),
            ('iBGPPeerCnt', 'asn == peerAsn', 'peer'),
            ('eBGPPeerCnt', 'asn != peerAsn', 'peer'),
            ('rrClientPeerCnt', 'rrclient.str.lower() == "true"', 'peer',
             'count'),
        ]

        self._gen_summarize_data()

        {self.ns[i].update({'activeAfiSafiCnt': afi_safi_count[i]})
         for i in self.ns.keys()}
        self.summary_row_order.append('activeAfiSafiCnt')

        self.summary_df['estdTime'] = humanize_timestamp(
            self.summary_df.estdTime,
            self.cfg.get('analyzer', {}).get('timezone', None))

        self.summary_df['estdTime'] = (
            self.summary_df['timestamp'] - self.summary_df['estdTime'])
        self.summary_df['estdTime'] = self.summary_df['estdTime'] \
                                          .apply(lambda x: x.round('s'))
        # Now come the BGP specific ones
        established = self.summary_df.query("state == 'Established'") \
            .groupby(by=['namespace'])

        uptime = established["estdTime"]
        rx_updates = established["updatesRx"]
        tx_updates = established["updatesTx"]
        self._add_stats_to_summary(uptime, 'upTimeStat')
        self._add_stats_to_summary(rx_updates, 'updatesRxStat')
        self._add_stats_to_summary(tx_updates, 'updatesTxStat')

        self.summary_row_order.extend(['upTimeStat', 'updatesRxStat',
                                       'updatesTxStat'])

        self._post_summarize()
        return self.ns_df.convert_dtypes()

    def _get_peer_matched_df(self, df) -> pd.DataFrame:
        """Get a BGP dataframe that also contains a session's matching peer"""

        if 'peerHostname' not in df.columns:
            return df

        # We have to separate out the Established and non Established entries
        # for the merge. Otherwise we end up with a mess
        df_1 = df[['namespace', 'hostname', 'vrf', 'peer', 'peerIP',
                   'updateSource']] \
            .drop_duplicates() \
            .reset_index(drop=True)
        df_2 = df[['namespace', 'hostname', 'vrf', 'updateSource']] \
            .drop_duplicates() \
            .reset_index(drop=True)

        mdf = df_1.merge(df_2,
                         left_on=['namespace', 'peerIP'],
                         right_on=['namespace', 'updateSource'],
                         suffixes=('', '_y')) \
            .drop_duplicates(subset=['namespace', 'hostname', 'vrf',
                                     'peerIP']) \
            .rename(columns={'hostname_y': 'peerHost'}) \
            .fillna(value={'peerHostname': '', 'peerHost': ''}) \
            .reset_index(drop=True)

        df = df.merge(mdf[['namespace', 'hostname', 'vrf', 'peer', 'peerHost']],
                      on=['namespace', 'hostname', 'vrf', 'peer'], how='left')

        df['peerHostname'] = np.where((df['peerHostname'] == '') &
                                      (df['state'] == "Established"),
                                      df['peerHost'],
                                      df['peerHostname'])
        df = df.fillna(value={'peerHostname': ''}) \
            .drop(columns=['peerHost'])

        for i in df.select_dtypes(include='category'):
            df[i].cat.add_categories('', inplace=True)

        return df

    def aver(self, **kwargs) -> pd.DataFrame:
        """BGP Assert"""

        assert_cols = ["namespace", "hostname", "vrf", "peer", "peerHostname",
                       "afi", "safi", "asn", "state", "peerAsn", "bfdStatus",
                       "reason", "notificnReason", "afisAdvOnly",
                       "afisRcvOnly", "peerIP", "updateSource"]

        kwargs.pop("columns", None)  # Loose whatever's passed
        status = kwargs.pop("status", 'all')

        df = self.get(columns=assert_cols, state='!dynamic', **kwargs)
        if 'error' in df:
            return df

        if df.empty:
            if status != "pass":
                df['assert'] = 'fail'
                df['assertReason'] = 'No data'
            return df

        df = self._get_peer_matched_df(df)
        if df.empty:
            if status != "pass":
                df['assert'] = 'fail'
                df['assertReason'] = 'No data'
            return df

        # We can get rid of sessions with duplicate peer info since we're
        # interested only in session info here
        df = df.drop_duplicates(
            subset=['namespace', 'hostname', 'vrf', 'peer'])
        df['assertReason'] = [[] for _ in range(len(df))]

        df['assertReason'] += df.apply(lambda x: ["asn mismatch"]
                                       if x['state'] != "Established" and
                                       (x['peerHostname'] and
                                        ((x["asn"] != x["peerAsn_y"]) or
                                         (x['asn_y'] != x['peerAsn'])))
                                       else [], axis=1)

        # Get list of peer IP addresses for peer not in Established state
        # Returning to performing checks even if we didn't get LLDP/Intf info
        df['assertReason'] += df.apply(
            lambda x: [f"{x['reason']}:{x['notificnReason']}"]
            if ((x['state'] != 'Established') and
                (x['reason'] and x['reason'] != 'None' and
                 x['reason'] != "No error"))
            else [],
            axis=1)

        df['assertReason'] += df.apply(
            lambda x: ['Matching BGP Peer not found']
            if x['state'] == 'NotEstd' and not x['assertReason']
            else [],  axis=1)

        df['assertReason'] += df.apply(
            lambda x: ['Not all Afi/Safis enabled']
            if x['afisAdvOnly'].any() or x['afisRcvOnly'].any() else [],
            axis=1)

        df['assert'] = df.apply(lambda x: 'pass'
                                if not len(x.assertReason) else 'fail',
                                axis=1)

        result = df[['namespace', 'hostname', 'vrf', 'peer', 'asn',
                     'peerAsn', 'state', 'peerHostname', 'assert',
                     'assertReason', 'timestamp']] \
            .explode(column="assertReason") \
            .fillna({'assertReason': '-'})

        if status == "fail":
            return result.query('assertReason != "-"')
        elif status == "pass":
            return result.query('assertReason == "-"')

        return result



from suzieq.sqobjects import *
from suzieq.gui.pages import *
from suzieq.gui.guiutils import (get_image_dir, display_title,
                                 get_main_session_by_id, SuzieqMainPages)
import streamlit as st
from streamlit import caching
from types import ModuleType
from collections import defaultdict
import base64


def build_pages():
    '''Build the pages and the corresponding functions to be called'''

    page_tbl = defaultdict(dict)
    module_list = globals()
    for key in module_list:
        if isinstance(module_list[key], ModuleType):
            if module_list[key].__package__ == 'suzieq.gui.pages':
                objlist = filter(
                    lambda x: x == "page_work" or x == "get_title",
                    dir(module_list[key]))
                page_name = None
                for obj in objlist:
                    if obj == 'get_title':
                        page_name = getattr(module_list[key], obj)()
                    else:
                        work_fn = getattr(module_list[key], obj)
                if page_name:
                    page_tbl[page_name] = work_fn

    return page_tbl


def build_sqobj_table() -> dict:
    '''Build available list of suzieq table objects'''

    sqobj_tables = {}
    module_list = globals()
    blacklisted_tables = ['path', 'topmem', 'topcpu', 'ifCounters', 'topology',
                          'time']
    for key in module_list:
        if isinstance(module_list[key], ModuleType):
            if key in blacklisted_tables:
                continue
            if module_list[key].__package__ == 'suzieq.sqobjects':
                objlist = list(filter(lambda x: x.endswith('Obj'),
                                      dir(module_list[key])))
                for obj in objlist:
                    sqobj_tables[key] = getattr(module_list[key], obj)

    return sqobj_tables


def apprun():
    '''The main application routine'''

    st.set_page_config(layout="wide", page_title="Suzieq")

    state = st.session_state
    for key, val in [('pages', None),
                     ('pagelist', []),
                     ('page', None),
                     ('search_text', ''),
                     ('sqobjs', {})]:
        if key not in state:
            state[key] = val

    if not state.pages:
        state.pages = build_pages()
        state.sqobjs = build_sqobj_table()
        # Hardcoding the order of these three
        state.pagelist = [SuzieqMainPages.STATUS.value,
                          SuzieqMainPages.XPLORE.value,
                          SuzieqMainPages.PATH.value,
                          SuzieqMainPages.SEARCH.value]
        for pg in state.pages:
            if pg not in state.pagelist:
                state.pagelist.append(pg)

    url_params = st.experimental_get_query_params()
    if url_params.get('page', ''):
        page = url_params['page']
        if isinstance(page, list):
            page = page[0]

        old_session_state = get_main_session_by_id(
            url_params.get('session', [''])[0])
        if page == "_Path_Debug_":
            state.pages[page](old_session_state, True)
            st.stop()
        elif page == "_Help_":
            state.pages[page](old_session_state, None)
            st.stop()
        if isinstance(page, list):
            page = page[0]
    else:
        page = None

    if page is None:
        if state.page:
            page = state.page
        else:
            page = state.pagelist[0]

    state.page = page

    page, search_text = display_title(page)

    if search_text != state.search_text:
        state.search_text = search_text
        state.page = page = 'Search'

    state.pages[page](state)


if __name__ == '__main__':
    apprun()



import argparse
from suzieq.cli.sqcmds import *
from suzieq.cli.sqcmds import context_commands
from suzieq.cli.sqcmds import sqcmds_all
from suzieq.cli.sq_nubia_context import NubiaSuzieqContext
from suzieq.cli.sq_nubia_statusbar import NubiaSuzieqStatusBar
from nubia import PluginInterface, CompletionDataSource
from nubia.internal.blackcmd import CommandBlacklist
from nubia.internal.cmdbase import AutoCommand


class NubiaSuzieqPlugin(PluginInterface):
    """
    The PluginInterface class is a way to customize nubia for every customer
    use case. It allowes custom argument validation, control over command
    loading, custom context objects, and much more.
    """

    def create_context(self):
        """
        Must create an object that inherits from `Context` parent class.
        The plugin can return a custom context but it has to inherit from the
        correct parent class.
        """
        return NubiaSuzieqContext()

    def validate_args(self, args):
        """
        This will be executed when starting nubia, the args passed is a
        dict-like object that contains the argparse result after parsing the
        command line arguments. The plugin can choose to update the context
        with the values, and/or decide to raise `ArgsValidationError` with
        the error message.
        """
        pass

    def get_commands(self):
        cmds = [AutoCommand(getattr(globals()[x], x))
                for x in sqcmds_all if not x.startswith(('_', 'ArgHelpClass'))]
        cmds.append(AutoCommand(context_commands.set_ctxt))
        cmds.append(AutoCommand(context_commands.clear_ctxt))
        return cmds

    def get_opts_parser(self, add_help=True):
        """
        Builds the ArgumentParser that will be passed to , use this to
        build your list of arguments that you want for your shell.
        """
        opts_parser = argparse.ArgumentParser(
            description="Suzieq CLI",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            add_help=add_help,
        )
        opts_parser.add_argument(
            "--config", "-c", default="", type=str, help="Configuration File"
        )
        opts_parser.add_argument(
            "--verbose",
            "-v",
            action="count",
            default=0,
            help="Increase verbosity, can be specified " "multiple times",
        )
        opts_parser.add_argument(
            "--stderr",
            "-s",
            action="store_true",
            default=True,
            help="By default the logging output goes to stderr "
            "Enable this feature to send it to a temporary logfile"
        )
        # we only support pandas now, so we don't want this option
        # opts_parser.add_argument(
        #    "--use-engine", "-e", help="Which analysis engine to use", default="pandas"
        # )
        return opts_parser

    def get_completion_datasource_for_global_argument(self, argument):
        if argument == "--use-engine":
            return ConfigEngineCompletionDataSource()
        return None

    def create_usage_logger(self, context):
        """
        Override this and return you own usage logger.
        Must be a subtype of UsageLoggerInterface.
        """
        return None

    def get_status_bar(self, context):
        """
        This returns the StatusBar object that handles the bottom status bar
        and the right-side per-line status
        """
        return NubiaSuzieqStatusBar(context)

    def getBlacklistPlugin(self):
        blacklister = CommandBlacklist()
        blacklister.add_blocked_command("topcpu")
        blacklister.add_blocked_command("topmem")
        return blacklister


class ConfigEngineCompletionDataSource(CompletionDataSource):
    def get_all(self):
        return ["pandas"]



import time
from datetime import timedelta
from nubia import command, argument
import pandas as pd

from suzieq.utils import humanize_timestamp
from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.bgp import BgpObj


@command("bgp", help="Act on BGP data")
class BgpCmd(SqCommand):
    """BGP protocol information"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        columns: str = "default",
        query_str: str = ' ',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=BgpObj,
        )

    def _clean_output(self, df) -> pd.DataFrame:
        """Make upTime look good"""
        if df.empty:
            return df

        if "estdTime" in df.columns:
            df['estdTime'] = df.estdTime.apply(
                lambda x: str(timedelta(milliseconds=int(x))))
        elif 'upTimeStat' in df.columns:
            df.loc['upTimeStat'] = df \
              .loc['upTimeStat'] \
              .map(lambda x: [str(timedelta(milliseconds=int(i))) for i in x])

        return df.dropna(how='any')

    @command("show")
    @argument("vrf", description="vrf name to qualify")
    @argument("peer", description="IP address, in quotes, or the interface name, of peer to qualify output")
    @argument("state", description="status of the session to match",
              choices=["Established", "NotEstd", "dynamic"])
    def show(self, state: str = "", vrf: str = '', peer: str = ''):
        """Show BGP info
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        if (self.columns != ['default'] and self.columns != ['*'] and
                'state' not in self.columns):
            addnl_fields = ['state']
        else:
            addnl_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname, columns=self.columns,
                                namespace=self.namespace, state=state,
                                addnl_fields=addnl_fields,
                                query_str=self.query_str,
                                vrf=vrf.split(), peer=peer.split()
                                )

        if 'estdTime' in df.columns and not df.empty:
            df['estdTime'] = humanize_timestamp(df.estdTime,
                                                self.cfg.get('analyzer', {})
                                                .get('timezone', None))

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @command("summarize", help="Provide summary info about BGP per namespace")
    def summarize(self):
        """
        Summarize bgp info
        """
        self._init_summarize()
        return self._post_summarize()

    @command("assert")
    @argument("vrf", description="Only assert BGP state in this VRF")
    @argument("status", description="Show only assert that matches this value",
              choices=["all", "fail", "pass"])
    def aver(self, vrf: str = "", status: str = "all") -> pd.DataFrame:
        """Assert BGP is functioning properly"""

        now = time.time()
        df = self._invoke_sqobj(self.sqobj.aver,
                                vrf=vrf.split(),
                                namespace=self.namespace,
                                status=status,
                                )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._assert_gen_output(df)




import time
import pandas as pd
from dataclasses import dataclass
from nubia import command, argument, context
from io import StringIO
import shutil
from prompt_toolkit import prompt
from suzieq.exceptions import UserQueryError
from natsort import natsort_keygen
import inspect
from colorama import Fore, Style


@dataclass
class ArgHelpClass(object):
    '''Class holding description and other params to display help
    This class exists to simplify displaying help for vars that we
    cannot access via the classargspec var
    '''
    description: str = ''


@argument(
    "engine",
    description="which analytical engine to use",
    choices=["pandas"],
)
@argument(
    "namespace", description="Space separated list of namespaces to qualify"
)
@argument("hostname", description="Space separated list of hostnames to qualify")
@argument(
    "start_time", description="Start of time window, try natural language spec"
)
@argument(
    "end_time", description="End of time window, try natural language spec "
)
@argument(
    "view",
    description="view all records or just the latest",
    choices=["all", "latest"],
)
@argument("columns", description="Space separated list of columns, * for all")
@argument(
    "format",
    description="select the pformat of the output",
    choices=["text", "json", "csv", "markdown"],
)
@argument(
    "query_str",
    description="Trailing blank terminated pandas query format to further filter the output.",
)
class SqCommand:
    """Base Command Class for use with all verbs"""

    def __init__(
            self,
            engine: str = "pandas",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",
            columns: str = "default",
            query_str: str = " ",
            sqobj=None,
    ) -> None:
        self.ctxt = context.get_context()
        self._cfg = self.ctxt.cfg
        self._schemas = self.ctxt.schemas
        self.format = format or "text"
        self.columns = columns.split()
        self.json_print_handler = None
        self._additional_help_vars = {}

        if query_str.count('"') % 2 != 0:
            # This happens because nubia strips off the trailing quote
            # if not followed by a blank
            query_str += '"'
        self.query_str = query_str.strip()

        if not isinstance(namespace, str):
            print('namespace must be a space separated list of strings')
            return
        if not isinstance(hostname, str):
            print('hostname must be a space separated list of strings')
            return
        if not isinstance(columns, str):
            print('columns must be a space separated list of strings')
            return

        if not namespace and self.ctxt.namespace:
            self.namespace = self.ctxt.namespace
        else:
            self.namespace = namespace.split()
        if not hostname and self.ctxt.hostname:
            self.hostname = self.ctxt.hostname
        else:
            self.hostname = hostname.split()

        if not start_time and self.ctxt.start_time:
            self.start_time = self.ctxt.start_time
        else:
            self.start_time = start_time

        if not end_time and self.ctxt.end_time:
            self.end_time = self.ctxt.end_time
        else:
            self.end_time = end_time

        self.view = view

        if not sqobj:
            raise AttributeError('mandatory parameter sqobj missing')

        self.sqobj = sqobj(context=self.ctxt,
                           hostname=self.hostname,
                           start_time=self.start_time,
                           end_time=self.end_time,
                           view=self.view,
                           namespace=self.namespace,
                           columns=self.columns)

    @property
    def cfg(self):
        return self._cfg

    @property
    def schemas(self):
        return self._schemas

    @command("summarize", help='produce a summarize of the data')
    def summarize(self):
        """Summarize relevant information about the table"""
        if not self._init_summarize():
            return self._gen_output(self.summarize_df)

        return self._post_summarize()

    @command("unique", help="find the list of unique items in a column")
    @argument("count", description="include count of times a value is seen",
              choices=['True'])
    def unique(self, count='', **kwargs):
        """Get all unique values associated with requested field"""
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.unique,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                count=count,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if 'error' in df.columns:
            return self._gen_output(df)

        if df.empty:
            return df

        if not count:
            return self._gen_output(df.sort_values(by=[self.columns[0]]),
                                    dont_strip_cols=True)
        else:
            return self._gen_output(
                df.sort_values(by=['numRows', self.columns[0]]),
                dont_strip_cols=True)

    @command("describe", help="describe the table and its fields")
    def describe(self):
        """Display the schema of the table

        Returns:
            [type]: 0 or error
        """
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.describe,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @command("top", help="find the top n values for a field")
    @argument("count", description="number of rows to return")
    @argument("what", description="integer field to get top values for")
    @argument("reverse", description="return bottom n values",
              choices=['True', 'False'])
    def top(self, count: int = 5, what: str = '', reverse: str = 'False',
            **kwargs) -> int:
        """Return the top n values for a field in a table

        Args:
            count (int, optional): The number of entries to return. Defaults to 5
            what (str, optional): Field name to use for largest/smallest val
            reverse (bool, optional): Reverse and return n smallest

        Returns:
            int: 0 or error code
        """
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.top,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                what=what, count=count,
                                reverse=eval(reverse),
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if 'error' in df.columns:
            return self._gen_output(df)

        if not df.empty:
            df = self.sqobj.humanize_fields(df)
            return self._gen_output(df.sort_values(by=[what], ascending=False),
                                    dont_strip_cols=True, sort=False)
        else:
            return self._gen_output(df)

    @command("help", help="show help for a command")
    @argument("command", description="command to show help for",
              choices=['show', 'unique', 'summarize', 'assert', 'describe',
                       'top', "find", "lpm"])
    def help(self, command: str = ''):
        """Show help for a command

        Args:
            command (str, optional): the name of the command. Defaults to 'show'.
        """
        if any(x for x in [self.namespace, self.hostname, self.view,
                           self.start_time, self.end_time, self.query_str]):
            print(Fore.RED + "Error: Only accepeted options is command")
            return
        if (self.columns != ["default"]) or (self.format != "text"):
            print(Fore.RED + "Error: Only accepeted options is command")
            return

        if not command:
            mbrs = inspect.getmembers(self)
            verbs = {x[0]: x[1] for x in mbrs
                     if inspect.ismethod(x[1]) and not x[0].startswith('_')}
            print(f"{self.sqobj.table}: " + Fore.CYAN + f"{self.__doc__}" +
                  Style.RESET_ALL)
            print("\nSupported verbs are: ")
            for verb in verbs:
                docstr = inspect.getdoc(verbs[verb])
                if docstr:
                    docstr = docstr.splitlines()[0]
                else:
                    docstr = ''
                verb = verb.replace('aver', 'assert')
                print(f" - {verb}: " + Fore.CYAN + f"{docstr}" +
                      Style.RESET_ALL)
            print(f"\nUse " + Fore.CYAN + f"{self.sqobj.table} help command=<verb>" +
                  Style.RESET_ALL + " for more details on that verb")
        else:
            self._do_help(self.sqobj.table, command)

    def _do_help(self, table: str, verb: str = 'show'):
        """Show help for a command

        Args:
            table (str): the table name
            verb (str, optional): the name of the command. Defaults to 'show'.
        """

        mbrs = inspect.getmembers(self)
        classargspec = [x[1] for x in mbrs
                        if x[0] == '__arguments_decorator_specs']
        if classargspec:
            classargspec = classargspec[0]
        verbs = [x[0] for x in mbrs
                 if inspect.ismethod(x[1]) and not x[0].startswith('_')]
        fnlist = {x[0]: x[1] for x in mbrs
                  if x[0] in verbs}
        newverb = verb.replace('assert', 'aver')
        if newverb in fnlist:
            fnargs = []

            fnmbrs = inspect.getmembers(fnlist[newverb])
            for elem in fnmbrs:
                if elem[0] == "__arguments_decorator_specs":
                    fnargs = elem[1]
                    break

            docstr = inspect.getdoc(fnlist[newverb])
            if docstr:
                docstr = docstr.splitlines()[0]
            else:
                docstr = ''
            print(f"{table} {verb}: " + Fore.CYAN +
                  f"{docstr}" + Style.RESET_ALL)
            print(Fore.YELLOW + '\nArguments:' + Style.RESET_ALL)
            if fnargs:
                classargspec.update(fnargs)
            classargspec.update(self._additional_help_vars)
            for arg in sorted(classargspec):
                print(f" - {arg}: " + Fore.CYAN +
                      f"{classargspec[arg].description}" + Style.RESET_ALL)
        else:
            print(f"No information about {table} {verb}")

    def _pager_print(self, df: pd.DataFrame) -> None:
        '''To support paging'''

        if isinstance(df, pd.DataFrame) and df.index.dtype == 'int64':
            df.reset_index(drop=True, inplace=True)

        if self.ctxt.pager:
            screen_lines = (shutil.get_terminal_size((80, 20)).lines - 2)
            bufio = StringIO()

            print(df, file=bufio)
            lines = bufio.getvalue().split('\n')
            curline = 0
            total_lines = len(lines)
            while curline < total_lines:
                print('\n'.join(lines[curline:curline+screen_lines]))
                curline += screen_lines
                if curline < total_lines:
                    try:
                        _ = prompt('Hit <ENTER> to continue, <CTRL-D> to quit')
                    except EOFError:
                        break
        else:
            print(df)

        return

    def _gen_output(self, df: pd.DataFrame, json_orient: str = "records",
                    dont_strip_cols: bool = False, sort: bool = True):

        if 'error' in df.columns:
            retcode = 1
            max_colwidth = None
            cols = df.columns
            is_error = True
        else:
            max_colwidth = self.ctxt.col_width
            retcode = 0
            if self.columns != ['default'] and self.columns != ['*']:
                cols = self.columns
            else:
                cols = df.columns
            is_error = False

        if dont_strip_cols or not all(item in df.columns for item in cols):
            cols = df.columns

        if self.format == 'json':
            if self.json_print_handler:
                print(df[cols].to_json(
                    default_handler=self.json_print_handler,
                    orient=json_orient))
            else:
                print(df[cols].to_json(orient=json_orient))
        elif self.format == 'csv':
            print(df[cols].to_csv())
        elif self.format == 'markdown':
            print(df[cols].to_markdown())
        elif self.format == 'devconfig' and self.sqobj.table == "devconfig":
            for row in df.itertuples():
                self._pager_print(row.config)
        else:
            with pd.option_context('precision', 3,
                                   'display.max_colwidth', max_colwidth,
                                   'display.max_rows', 256):
                if df.empty:
                    print(df)
                elif sort:
                    if is_error:
                        print(df[cols])
                    else:
                        sort_fields = [x for x in self.sqobj._sort_fields
                                       if x in df.columns and x in cols]
                        if sort_fields:
                            self._pager_print(
                                df[cols].sort_values(by=sort_fields,
                                                     key=natsort_keygen()))
                        else:
                            self._pager_print(df[cols])
                else:
                    self._pager_print(df[cols])

        return retcode

    def _assert_gen_output(self, df):
        if df.empty:
            result = 0
        elif df.columns.to_list() == ['error']:
            result = 1
        elif df.loc[df['assert'] != "pass"].empty:
            result = 0
        else:
            result = -1

        self._gen_output(df)
        if self.format == 'text':
            if result == 0:
                print("Assert passed")
            else:
                print("Assert failed")
        return result

    def _invoke_sqobj(self, fn, **kwargs) -> pd.DataFrame:

        try:
            df = fn(**kwargs)
        except Exception as ex:
            if isinstance(ex, UserQueryError):
                df = pd.DataFrame({'error': [f'ERROR: UserQueryError: {ex}']})
            else:
                df = pd.DataFrame({'error': [f'ERROR: {ex}']})

        return df

    def _init_summarize(self):
        self.now = time.time()

        self.summarize_df = self._invoke_sqobj(
            self.sqobj.summarize, namespace=self.namespace,
            hostname=self.hostname, query_str=self.query_str
        )
        return 'error' not in self.summarize_df

    def _post_summarize(self):
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - self.now)
        return self._gen_output(self.summarize_df, json_orient='columns')



import re
import ast
import logging
import operator as op


def _stepdown_rest(entry) -> list:
    '''move the elements one level down into the existing JSON hierarchy.

    if `entry` has the structure {'vrfs': {'routes'...} on invocation,
    it'll be {'routes'...} on exit. It makes no sense to move down them
    hierarchy when you have only a list.
    '''
    if isinstance(entry["rest"], dict):
        keylist = list(entry["rest"].keys())
        data = entry["rest"]
        entry = []
        for key in keylist:
            entry += [{"rest": data[key]}]
    else:
        entry = [entry]

    return entry


def parse_subtree(subflds, tmpval, maybe_list, defval):

    value = []
    useval = False

    for i, subfld in enumerate(subflds):
        subfld = subfld.strip()
        if ':_sqstore' in subfld:
            storekey = True
            subfld = '*'
        else:
            storekey = False
        if subfld == "*":
            tmp = tmpval
            for subele in tmp:
                if storekey:
                    storeval = subele
                for ele in subflds[i+1:]:
                    if isinstance(tmp, list):
                        subele = subele.get(ele, None)
                    else:
                        subele = tmp[subele].get(ele, None)
                    if subele is None:
                        tmpval = None
                        break
                if subele:
                    if storekey:
                        value.append(storeval)
                    else:
                        value.append(subele)
                    useval = True
            break
        elif subfld.startswith('['):
            # handle specific array index or dict key
            # This additional handling of * is because of JUNOS
            # interface address which is many levels deep and mixes
            # IPv4/v6 address in a way that we can't entirely fix
            # Post processing cleanup has to handle that part
            if subfld.endswith('?'):
                # This was added thanks to NXOS which provides
                # nexthops as a list if ECMP, else a dictionary
                if type(tmpval) != list:
                    continue
                subfld = subfld[:-1]  # lose the final "?" char
            if subfld == '[*]':
                for item in tmpval:
                    newval = parse_subtree(subflds[i+1:], item, maybe_list,
                                           defval)
                    if newval is not None:
                        if isinstance(newval, list):
                            value += newval
                        else:
                            value.append(newval)
                useval = True
                break
            elif tmpval:
                tmpval = tmpval[eval_expr(subfld)]
        elif isinstance(tmpval, dict):
            tmpval = tmpval.get(subfld, None)
        else:
            tmpval = None

        if not tmpval:
            break

    if useval:
        if value:
            return value
        else:
            return defval
    else:
        return tmpval


def cons_recs_from_json_template(tmplt_str, in_data):
    ''' Return an array of records given the template and input data.

    This uses an XPATH-like template string to create a list of records
    matching the template. It also normalizes the key fields so that we
    can create records with keys that are agnostic of the source.

    I could not use a ready-made library like jsonpath because of the
    difficulty in handling normalization and how jsonpath returns the
    result. For example, if I have 3 route records, one with 2 nexthop IPs,
    one with a single nexthop IP and one without a nexthop IP, jsonpath
    returns the data as a single flat list of nexthop IPs without a hint of
    figuring out which route the nexthops correspond to. We also support
    some amount of additional processing on the extracted fields such as
    the basic 4 arithmetic operations and specifying a default or
    substitute.
    '''
    result = []
    # Find prefix string
    try:
        ppos = re.search(r'/\[\s+', tmplt_str).start()
    except AttributeError:
        ppos = tmplt_str.index('[')

    # templates have a structure with a leading hierarchy traversal
    # followed by the fields for each record within that hierarchy.
    # One example is: vrfs/*:vrf/routes/*:prefix/[... where '[' marks
    # the start of the template to extract the values from the route
    # records (prefix) across all the VRFs(vrf). We break up this
    # processing into two parts, one before we reach  the inner record
    # (before '[') and the other after.
    #
    # Before we enter the inner records, we flatten the hierarchy by
    # creating as many records as necessary with the container values
    # filled into each record as a separate key. So with the above
    # template, we transform the route records to carry the vrf and the
    # prefix as fields of each record. Thus, when we have 2 VRFs and 10
    # routes in the first VRF and 6 routes in the second VRF, before
    # we're done with the processing of the prefix hierarchy, the result
    # has 16 records (10+6 routes) with the VRF and prefix values contained
    # in each record.
    #
    # We flatten because that is how pandas (and pyarrow) can process the
    # data best and queries can be made simple. The only nested structure
    # we allow is a list in the innermost fields. Thus, the list of nexthop
    # IP addresses associated with a route or the list of IP addresses
    # associated with an interface are allowed, but not a tuple consisting
    # of the nexthopIP and oif as a single entry.
    data = in_data
    nokeys = True
    try:
        pos = tmplt_str.index("/")
    except ValueError:
        ppos = 0                # completely flat JSON struct
    while ppos > 0:
        xstr = tmplt_str[0:pos]

        if ":" not in xstr:
            if not result:
                if xstr != '*' and xstr != '*?':
                    if not data or not data.get(xstr, None):
                        # Some outputs contain just the main key with a null
                        # body such as ospfNbr from EOS: {'vrfs': {}}.
                        logging.info(
                            f"Unnatural return from svcparser. xstr is {xstr}. \
                            Result is {result}")
                        return cleanup_and_return(result)
                    result = [{"rest": data[xstr]}]
                else:
                    if isinstance(data, dict):
                        result = [{"rest": []}]
                        for key in data.keys():
                            result[0]["rest"].append(data[key])
                    else:
                        result = [{"rest": data}]
            else:
                if xstr != "*" and xstr != '*?':
                    # Handle xstr being a specific array index or dict key
                    if xstr.startswith('['):
                        if len(result[0]['rest']):
                            result[0]["rest"] = result[0]["rest"][eval_expr(
                                xstr)]
                        else:
                            # Handling the JUNOS EVPN pfx DB entry
                            logging.info(
                                f'Unnatural return from svcparser. '
                                f'xstr is {xstr}. Result is {result}')
                            return []

                    else:
                        tmpval = []
                        for ele in result:
                            # EOS routes case: vrfs/*:vrf/routes/*:prefix
                            # Otherwise there's usually one element here
                            if isinstance(ele["rest"], list) and ele['rest']:
                                # We need to check for empty rest because in 
                                # older NXOS, where there are OSPF neighbors 
                                # defined in a VRF, we can end up in this 
                                # if clause and miss out on good data thats
                                # actually present in the else clause of this. 
                                for subidx, subele in enumerate(ele["rest"]):
                                    if xstr in subele:
                                        if nokeys:
                                            tmpval.append(
                                                {"rest": subele[xstr]})
                                        else:
                                            ele['rest'][subidx] = subele[xstr]
                                if not nokeys:
                                    if len(ele['rest']) == 1:
                                        ele['rest'] = ele['rest'][0]

                                    tmpval.append(ele)
                            else:
                                if xstr in ele['rest']:
                                    ele["rest"] = ele["rest"][xstr]

                        if tmpval:
                            result = tmpval
                else:
                    if (xstr == '*?'):
                        # Massaging the format to handle cases like NXOS that
                        # provide dict when there's a single element and a list
                        # if there's more than one element
                        tmpres = []
                        for item in result:
                            if not isinstance(item, list):
                                if not isinstance(item['rest'], list):
                                    item['rest'] = [item['rest']]
                                tmpres.append([item])
                        result = tmpres
                    else:
                        result = list(map(_stepdown_rest, result))
                    if len(result) == 1:
                        result = result[0]
                    else:
                        tmpres = result[0]
                        if tmpres[0].keys() != set(['rest']):
                            # Handle output like EOS OSPF output with the
                            # starting string:
                            # vrfs/*:vrf/instList/*:instance/ospfNeighborEntries/*/[
                            # Move the keys into each element of the lists
                            # we need to pop out these keys at the very end
                            for entry in result:
                                elekeys = entry[0].keys() - set(['rest'])
                                if not elekeys:
                                    # this happens when NXOS routes returns
                                    # half-baked data when there are no routes
                                    # in a VRF.
                                    continue
                                for rstentry in entry[0]['rest']:
                                    rstentry['sq-addnl-keys'] = []
                                    for key in elekeys:
                                        rstentry['sq-addnl-keys'].append({
                                            key: entry[0][key]})
                                for key in elekeys:
                                    del entry[0][key]
                                # We should only have 'rest' entries now
                            nokeys = True  # We've moved all the external keys in

                        # Handle the output of the likes of EOS' BGP with
                        # starting string: 'vrfs/*/peerList/*/[ by
                        # merging the rest lists
                        for entry in result[1:]:
                            tmpres[0]['rest'].extend(entry[0]['rest'])
                        result = tmpres

            tmplt_str = tmplt_str[pos + 1:]
            if re.match(r'^\[\s+"', tmplt_str):
                ppos = 0
                continue
            try:
                pos = tmplt_str.index("/")
            except ValueError:
                # its ppossible the JSON data is entirely flat
                ppos = 0
                continue
            ppos -= pos
            continue

        # handle one level of nesting to deal with Junos route JSON, NXOS route
        # and many others that have an interesting field in parallel with
        # the rest of the useful data
        *lval, rval = xstr.split(":")
        if "|" in rval:
            rval, nxtfld = rval.split('|')
        else:
            nxtfld = None
        nokeys = False
        ks = [lval[0]]
        tmpres = []
        intres = []
        if result:
            for ele in result:
                if lval[0] == "*":
                    if isinstance(ele['rest'], dict):
                        if nxtfld:
                            intres = [{rval: ele['rest'].get(lval[1], ''),
                                       "rest": ele['rest'].get(nxtfld, [])}]
                        else:
                            ks = list(ele["rest"].keys())

                            intres = [{rval: x,
                                       "rest": ele["rest"][x]}
                                      for x in ks]
                    else:
                        if nxtfld:
                            intres = [{rval: x.get(lval[1], ''),
                                       "rest": x.get(nxtfld, [])}
                                      for x in ele["rest"]]
                        else:
                            intres = [{rval: x.get(lval[1], ''), "rest": x}
                                      for x in ele["rest"]]

                for oldkey in ele.keys():
                    if oldkey == "rest":
                        continue
                    for newele in intres:
                        newele.update({oldkey: ele[oldkey]})
                tmpres += intres
            result = tmpres
        else:
            if lval == ["*"]:
                ks = list(data.keys())

            result = [{rval: x,
                       "rest": data[x]} for x in ks]

        tmplt_str = tmplt_str[pos + 1:]
        try:
            # handle EOS' ospfIf output
            pos = tmplt_str.index("/")
        except ValueError:
            pos = ppos
        ppos -= pos

    # Now for the rest of the fields
    # if we only have 'rest' as the key, break out into individual mbrs
    if nokeys:
        if not result:
            result = [{"rest": data}]
        elif len(result) == 1:
            if isinstance(result[0], list):
                tmpval = []
                for x in result[0]["rest"]:
                    tmpval.append({"rest": x})
                result = tmpval

    # In some cases such as FRR's BGP, you need to eliminate elements which
    # have no useful 'rest' field, for example the elements with vrfId and
    # vrfName. If at this point, you have nn element in result with rest
    # that is not a list or a dict, remove it

    result = list(filter(lambda x: isinstance(x["rest"], list) or
                         isinstance(x["rest"], dict),
                         result))

    # At this point, we're expecting result to contain a list where each
    # entry contains the "rest' key with the fields from which further data
    # is to be extracted, and a series of keys which represent what has already
    # been extracted from the header string. If the format of result isn't
    # this, fix it
    if len(result) == 1 and isinstance(result[0]['rest'], list):
        tmpres = []
        entry = result[0]
        elekeys = entry.keys() - set(['rest'])
        for elem in entry['rest']:
            newentry = {}
            [newentry.update({x: entry[x]}) for x in elekeys]
            newentry['rest'] = elem
            tmpres.append(newentry)
        result = tmpres

    # The if handles cases of flat JSON data such as evpnVni
    if tmplt_str.startswith('/['):
        tmplt_str = tmplt_str[2:-1]
    else:
        tmplt_str = tmplt_str[1:][:-1]         # eliminate'[', ']'
    for selem in tmplt_str.split(","):
        per_entry_defval = False
        def_val = None
        # every element here MUST have the form lval:rval
        selem = selem.replace('"', '').strip()
        if not selem:
            # dealing with trailing "."
            continue

        try:
            lval, rval = selem.split(": ")
        except ValueError:
            logging.error(f"Unable to parse JSON field entry {selem}")
            continue

        # Process default value processing of the form <key>?|<def_val> or
        # <key>?<expected_val>|<def_val>
        op = None
        if "?" in rval:
            rval, op = rval.split("?")
            exp_val, def_val = op.split("|")

            # Handle the case that the values are not strings
            if def_val.isdigit():
                def_val = int(def_val)
            elif def_val:
                if result and def_val not in result[0]:
                    # handle array indices, such as [] for default
                    # If the field to be init is a prev val, then handle this
                    # in the loop for x below as its different for each entry
                    # an example of such an entry is:
                    # "advertisedAndReceived: v4Enabled?|v4Enabled"
                    # which means if advertisedAndReceived is not found in this
                    # iteration, retain the previous value. This is useful when
                    # handling minor changes in JSON output such as one version
                    # with a key 'Advertised And Received' changing to
                    # 'advertisedAndReceived' in the next version of output
                    try:
                        adef_val = ast.literal_eval(def_val)
                        def_val = adef_val
                    except ValueError:
                        pass
                else:
                    per_entry_defval = True

        # Process for every element in result so far
        # Handles entries such as "vias/*/nexthopIps" and returns
        # a list of all nexthopIps.
        for x in result:

            loopdef_val = def_val
            if per_entry_defval and def_val is not None:
                if def_val in x:
                    loopdef_val = x[def_val]
                else:
                    loopdef_val = ''  # this shouldn't happen

            if "/" in lval:
                subflds = lval.split("/")
                tmpval = x["rest"]
                value = None
                if any(x in subflds for x in ["*", "*?", "[*]?", "[*]",
                                              '*:_sqstore']):
                    # Returning a list is the only supported option for now
                    value = []
                    maybe_list = True
                else:
                    maybe_list = False

                value = parse_subtree(
                    subflds, tmpval, maybe_list, loopdef_val)
                subflds = []
            else:
                if isinstance(x['rest'], dict):
                    value = x["rest"].get(lval.strip(), None)

            if op:
                if exp_val and value != exp_val:
                    value = loopdef_val
                elif not exp_val:
                    if (isinstance(value, list) and not
                            any(x or (x == loopdef_val) for x in value)):
                        if not isinstance(loopdef_val, list) and maybe_list:
                            if loopdef_val == '':
                                value = []
                            else:
                                value = [loopdef_val]
                        else:
                            value = loopdef_val
                    elif not value:
                        value = loopdef_val

            # Handle any operation on string
            rval = rval.strip()
            rval1 = re.split(r"([+/*-])", rval)
            if len(rval1) > 1:
                iop = rval1[1]
                if value is not None:
                    if rval1[0] in x:
                        value = eval_expr(f'{value}{iop}{x[rval1[0]]}')
                    else:
                        value = eval_expr(f'{value}{iop}{rval1[2]}')
                try:
                    x.update({rval1[0]: value})
                except ValueError:
                    x.update({rval1[0]: value})
                continue

            if (isinstance(value, str) and value.startswith('"') and
                    value.endswith('"')):
                # Strip leading and trailing quotes from string
                x.update({rval: value[1:-1]})
            else:
                x.update({rval: value})

    return cleanup_and_return(result)


def cleanup_and_return(result):
    for entry in result:
        # To handle entries like EOS' ospfNbr, we had saved the multiple keys
        # extracted from the initial part of the JSON string and saved it in
        # sq-addnl-keys. Now pop them out into the appropriate spot before
        # eliminating the 'rest' field which contains the unused fields
        rest = entry.pop('rest', {})
        if 'sq-addnl-keys' in rest:
            for elem in rest['sq-addnl-keys']:
                entry.update(elem)

    return result


def eval_expr(expr):
    """Evaluate numerical expression without eval or other packages"""
    return num_eval(ast.parse(expr, mode='eval').body)


def num_eval(node):
    # supported arithmetic operators
    operators = {ast.Add: op.add, ast.Sub: op.sub, ast.Mult: op.mul,
                 ast.Div: op.truediv, ast.Pow: op.pow}

    if isinstance(node, ast.Num):  # <number>
        return node.n
    elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return operators[type(node.op)](num_eval(node.left),
                                        num_eval(node.right))
    elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return operators[type(node.op)](num_eval(node.operand))
    elif isinstance(node, ast.List):
        return node.col_offset
    else:
        raise TypeError(node)



import logging
import typing
import pandas as pd

from suzieq.utils import load_sq_config, Schema, SchemaForTable
from suzieq.engines import get_sqengine


class SqContext(object):

    def __init__(self, engine, config_file=None):
        self.cfg = load_sq_config(config_file=config_file)

        self.schemas = Schema(self.cfg['schema-directory'])

        self.namespace = ''
        self.hostname = ''
        self.start_time = ''
        self.end_time = ''
        self.exec_time = ''
        self.engine = engine
        self.view = ''
        self.sort_fields = []


class SqObject(object):

    def __init__(self, engine_name: str = 'pandas',
                 hostname: typing.List[str] = [],
                 start_time: str = '', end_time: str = '',
                 view: str = '', namespace: typing.List[str] = [],
                 columns: typing.List[str] = ['default'],
                 context=None, table: str = '', config_file=None) -> None:

        if context is None:
            self.ctxt = SqContext(engine_name, config_file)
        else:
            self.ctxt = context
            if not self.ctxt:
                self.ctxt = SqContext(engine_name)

        self._cfg = self.ctxt.cfg
        self._schema = SchemaForTable(table, self.ctxt.schemas)
        self._table = table
        self._sort_fields = self._schema.key_fields()
        self._convert_args = {}

        if not namespace and self.ctxt.namespace:
            self.namespace = self.ctxt.namespace
        else:
            self.namespace = namespace
        if not hostname and self.ctxt.hostname:
            self.hostname = self.ctxt.hostname
        else:
            self.hostname = hostname

        if not start_time and self.ctxt.start_time:
            self.start_time = self.ctxt.start_time
        else:
            self.start_time = start_time

        if not end_time and self.ctxt.end_time:
            self.end_time = self.ctxt.end_time
        else:
            self.end_time = end_time

        if not view and self.ctxt.view:
            view = self.ctxt.view

        if self.start_time and self.end_time and not view:
            self.view = 'all'
        else:
            self.view = view or 'latest'

        self.columns = columns

        if engine_name and engine_name != '':
            self.engine = get_sqengine(engine_name,
                                       self._table)(self._table, self)
        elif self.ctxt.engine:
            self.engine = get_sqengine(self.ctxt.engine,
                                       self._table)(self._table, self)

        if not self.engine:
            raise ValueError('Unknown analysis engine')

        self._addnl_filter = None
        self._addnl_fields = []
        self._valid_get_args = None
        self._valid_assert_args = None
        self._valid_arg_vals = None
        self._valid_find_args = None

    @property
    def all_schemas(self):
        return self.ctxt.schemas

    @property
    def schema(self):
        return self._schema

    @property
    def cfg(self):
        return self._cfg

    @property
    def table(self):
        return self._table

    def _check_input_for_valid_args(self, good_arg_list, **kwargs,):
        if not good_arg_list:
            return

        # add standard args that are always
        good_arg_list = good_arg_list + (['namespace', 'addnl_fields'])

        for arg in kwargs.keys():
            if arg not in good_arg_list:
                raise AttributeError(
                    f"argument {arg} not supported for this command")

    def _check_input_for_valid_vals(self, good_arg_val_list, **kwargs):
        '''Check if the input is valid for the arg, if possible'''

        if not good_arg_val_list:
            return

        for arg in kwargs.keys():
            if arg in good_arg_val_list:
                if kwargs[arg] not in good_arg_val_list[arg]:
                    raise AttributeError(
                        f"invalid value {kwargs[arg]} for argument {arg}")

    def validate_get_input(self, **kwargs):
        self._check_input_for_valid_args(
            self._valid_get_args + ['columns'], **kwargs)
        self._check_input_for_valid_vals(self._valid_arg_vals, **kwargs)

    def validate_assert_input(self, **kwargs):
        self._check_input_for_valid_args(self._valid_assert_args, **kwargs)

    def validate_columns(self, columns: typing.List[str]) -> bool:
        """Validate that the provided columns are valid for the table

        Args:
            columns (List[str]): list of columns

        Returns:
            bool: True if columns are valid
        Raises:
            ValueError: if columns are invalid
        """

        if columns == ['default'] or columns == ['*']:
            return True

        table_schema = SchemaForTable(self._table, self.all_schemas)
        invalid_columns = [x for x in columns if x not in table_schema.fields]
        if invalid_columns:
            raise ValueError(f"Invalid columns specified: {invalid_columns}")

    def get(self, **kwargs) -> pd.DataFrame:

        if not self._table:
            raise NotImplementedError

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        if self._addnl_filter:
            kwargs['add_filter'] = self._addnl_filter

        # This raises exceptions if it fails
        try:
            self.validate_get_input(**kwargs)
        except Exception as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        if 'columns' not in kwargs:
            kwargs['columns'] = self.columns or ['default']

        # This raises ValueError if it fails
        self.validate_columns(kwargs.get('columns', []))

        for k in self._convert_args:
            v = kwargs.get(k, None)
            if v:
                val = kwargs[k]
                newval = []
                if isinstance(val, list):
                    for ele in val:
                        ele = self._convert_args[k](ele)
                        newval.append(ele)
                    kwargs[k] = newval
                elif isinstance(val, str):
                    kwargs[k] = self._convert_args[k](val)

        return self.engine.get(**kwargs)

    def summarize(self, namespace=[], hostname=[],
                  query_str='') -> pd.DataFrame:
        if self.columns != ["default"]:
            self.summarize_df = pd.DataFrame(
                {'error': ['ERROR: You cannot specify columns with summarize']})
            return self.summarize_df
        if not self._table:
            raise NotImplementedError

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        return self.engine.summarize(namespace=namespace, hostname=hostname,
                                     query_str=query_str)

    def unique(self, **kwargs) -> pd.DataFrame:
        if not self._table:
            raise NotImplementedError

        if not self.ctxt.engine:
            raise AttributeError('No analysis engine specified')

        columns = kwargs.pop('columns', self.columns)
        column = columns[0]

        if columns is None or columns == ['default']:
            raise ValueError('Must specify columns with unique')

        if len(columns) > 1:
            raise ValueError('Specify a single column with unique')

        # This raises ValueError if it fails
        self.validate_columns(columns)
        return self.engine.unique(**kwargs, columns=columns)

    def aver(self, **kwargs):
        raise NotImplementedError

    def top(self, what: str = '', count: int = 5, reverse: bool = False,
            **kwargs) -> pd.DataFrame:
        """Get the list of top/bottom entries of "what" field"""

        columns = kwargs.get('columns', ['default'])
        # This raises ValueError if it fails
        self.validate_columns(columns)

        if not what:
            raise ValueError('Must specify what field to get top for')
        # if self._valid_get_args:
        #     self._valid_get_args += ['what', 'n', 'reverse']
        # This raises exceptions if it fails
        try:
            self.validate_get_input(**kwargs)
        except Exception as error:
            df = pd.DataFrame({'error': [f'{error}']})
            return df

        # This raises ValueError if it fails
        table_schema = SchemaForTable(self._table, self.all_schemas)
        if not self._field_exists(table_schema, what):
            raise ValueError(
                f"Field {what} does not exist in table {self.table}")

        columns = table_schema.get_display_fields(columns)

        ftype = table_schema.field(what).get('type', 'str')
        if ftype not in ['long', 'double', 'float', 'int', 'timestamp',
                         'timedelta64[s]']:
            return pd.DataFrame({'error': [f'{what} not numeric; top can be used with numeric fields only']})

        if what not in columns:
            self._addnl_fields.append(what)

        return self.engine.top(what=what, count=count, reverse=reverse, **kwargs)

    def describe(self, **kwargs):
        """Describes the fields for a given table"""

        table = kwargs.get('table', self.table)

        try:
            sch = SchemaForTable(table, self.all_schemas)
        except ValueError:
            sch = None
        if not sch:
            df = pd.DataFrame(
                {'error': [f'ERROR: incorrect table name {table}']})
            return df

        entries = [{'name': x['name'], 'type': x['type'],
                    'key': x.get('key', ''),
                    'display': x.get('display', ''),
                    'description': x.get('description', '')}
                   for x in sch.get_raw_schema()]
        df = pd.DataFrame.from_dict(entries).sort_values('name')

        return df

    def get_table_info(self, table: str, **kwargs) -> pd.DataFrame:
        """Get some basic stats about the table from the database

        Args:
            table (str): The table to get stats for

        Returns:
            pd.DataFrame: A dataframe with the stats
        """
        # This raises ValueError if it fails
        self.validate_columns(kwargs.get('columns', ['default']))

        return self.engine.get_table_info(table, **kwargs)

    def humanize_fields(self, df: pd.DataFrame, subset=None) -> pd.DataFrame:
        '''Humanize the fields for human consumption.

        Individual classes will implement the right transofmations. This
        routine is just a placeholder for all those with nothing to modify.
        '''
        return df

    def _field_exists(self, table_schema: SchemaForTable, field: str) -> bool:
        """Check if a field exists in the schema

        Args:
            table_schema (SchemaForTable): The schema for the table
            field (str): the field name we're checking for

        Returns:
            bool: True if the field exists, False otherwise
        """
        return table_schema.field(field)



import time
import re
from types import resolve_bases
from nubia import command, argument
import pandas as pd

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.network import NetworkObj


@command("network", help="Act on network-wide data")
class NetworkCmd(SqCommand):
    """Overall network information such as namespaces present etc."""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",
            query_str: str = " ",
            columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=NetworkObj,
        )

    @command("show", help="Show device information")
    @argument("model", description="models to filter with")
    @argument("os", description='NOS to filter with')
    @argument('vendor', description='vendor to filter with')
    @argument('version', description='NOS version to filter with')
    def show(self, os: str = "", vendor: str = "", model: str = "",
             version: str = "") -> int:
        """
        Show network info
        """

        now = time.time()

        df = self.sqobj.get(namespace=self.namespace, os=os.split(),
                            vendor=vendor.split(), model=model.split(),
                            version=version,
                            hostname=self.hostname)

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._gen_output(df)

    @command("find", help="find an address, bgp session by asn, etc")
    @argument("address", type=str, description="IP/MAC address to find")
    @argument("vrf", type=str,
              description="Find within this VRF, used for IP addr")
    @argument("vlan", type=str,
              description="Find MAC within this VLAN")
    @argument("asn", type=str, description="Autonomous system number")
    @argument("resolve_bond", type=bool, description="Resolve the bond")
    def find(self, address: str = '', asn: str = '', vrf: str = '',
             vlan: str = '', resolve_bond: bool = False):
        """Find a network object such as an IP address, BGP session, etc.
        """
        now = time.time()

        df = self.sqobj.find(
            namespace=self.namespace,
            hostname=self.hostname,
            address=address.split(),
            asn=asn.split(),
            vlan=vlan,
            vrf=vrf,
            resolve_bond=resolve_bond,
        )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._gen_output(df)




from typing import Callable, Union
import pandas as pd
from pandas.core.computation.ops import UndefinedVariableError


def get_migrate_fn(table_name: str, from_vers: str,
                   to_vers: str) -> Union[Callable, None]:
    """Return a migration function if one is present for the table specified

    :param table_name: str, Name of the table for which we need the converter
    :param from_vers: str, Version number from which the conversion is needed
    :param to_vers: str, Version number to which the conversion is needed
    :returns: Routine to invoke to convert the data, or None
    :rtype: Callable or None

    """
    conversion_dict = {
        'bgp-1.0-2.0': _convert_bgp_vers_1_to_2
    }

    return conversion_dict.get(f'{table_name}-{from_vers}-{to_vers}', None)


def _convert_bgp_vers_1_to_2(df: pd.DataFrame) -> pd.DataFrame:
    """Convert BGP schema from version 1.0 to version 2.0

    The dataframe MUST contain the sqvers column

    """

    def set_community(x):
        communities = []
        if (x.afi != 'l2vpn'):
            return communities

        if x.evpnSendCommunity == 'extendedAndStandard':
            communities = ['standard', 'extended']
        elif x.evpnSendCommunity == 'extended':
            communities = ['extended']
        elif x.evpnSendCommunity == 'standard':
            communities = ['standard']
        return communities

    converted_df = pd.DataFrame()

    for pfx in ['v4', 'v6', 'evpn']:
        try:
            newdf = df.query(f'sqvers == "1.0" and {pfx}Enabled').reset_index()
        except UndefinedVariableError:
            newdf = pd.DataFrame()
        if not newdf.empty:
            if pfx == 'evpn':
                newdf['afi'] = 'l2vpn'
                newdf['safi'] = 'evpn'
            else:
                newdf['safi'] = 'unicast'
                if pfx == 'v4':
                    newdf['afi'] = 'ipv4'
                else:
                    newdf['afi'] = 'ipv6'

            newdf = newdf.rename(columns={
                f'{pfx}PfxRx': 'pfxRx',
                f'{pfx}PfxTx': 'pfxTx',
                f'{pfx}IngressRmap': 'ingressRmap',
                f'{pfx}EgressRmap': 'egressRmap',
                f'{pfx}defaultsent': 'defOriginate',
            })

            newdf['afisAdvOnly'] = [[] for _ in range(len(newdf))]
            newdf['afisRcvOnly'] = [[] for _ in range(len(newdf))]
            newdf['communityTypes'] = [[] for _ in range(len(newdf))]
            converted_df = pd.concat([converted_df, newdf])

    if not converted_df.empty:
        converted_df['afisAdvOnly'] += converted_df.apply(
            lambda x: ['ipv4 unicast']
            if (x.v4Advertised and not x.v4Received) else [], axis=1)
        converted_df['afisAdvOnly'] += converted_df.apply(
            lambda x: ['ipv6 unicast']
            if (x.v6Advertised and not x.v6Received) else [], axis=1)
        converted_df['afisAdvOnly'] += converted_df.apply(
            lambda x: ['l2vpn evpn']
            if (x.evpnAdvertised and not x.evpnReceived) else [], axis=1)

        converted_df['afisRcvOnly'] += converted_df.apply(
            lambda x: ['ipv4 unicast']
            if (not x.v4Advertised and x.v4Received) else [], axis=1)
        converted_df['afisRcvOnly'] += converted_df.apply(
            lambda x: ['ipv6 unicast']
            if (not x.v6Advertised and x.v6Received) else [], axis=1)
        converted_df['afisRcvOnly'] += converted_df.apply(
            lambda x: ['l2vpn evpn']
            if (not x.evpnAdvertised and x.evpnReceived) else [], axis=1)

        converted_df['communityTypes'] += converted_df.apply(set_community,
                                                             axis=1)

        converted_df['sqvers'] = '2.0'

        unconverted_df = df.query('sqvers != "1.0"').reset_index()
        final_df = pd.concat([converted_df, unconverted_df])

        return final_df.reset_index(drop=True)
    else:
        return df



from collections import defaultdict
from .engineobj import SqPandasEngine
import pandas as pd
import numpy as np
from typing import Tuple
from ipaddress import ip_address, ip_network


class RoutesObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'routes'

    def _cons_addnl_fields(self, columns: list,
                           addnl_fields: list) -> Tuple[list, list]:
        '''get all the additional columns we need'''

        drop_cols = []

        if columns == ['default']:
            addnl_fields.append('metric')
            drop_cols += ['metric']
        elif columns != ['*'] and 'metric' not in columns:
            addnl_fields.append('metric')
            drop_cols += ['metric']

            if 'ipvers' not in columns:
                addnl_fields.append('ipvers')
                drop_cols += ['ipvers']

        return addnl_fields, drop_cols

    def get(self, **kwargs):

        prefixlen = kwargs.pop('prefixlen', '')
        prefix = kwargs.pop('prefix', [])
        ipvers = kwargs.pop('ipvers', '')
        addnl_fields = kwargs.pop('addnl_fields', [])
        user_query = kwargs.pop('query_str', '')

        columns = kwargs.get('columns', ['default'])
        addnl_fields, drop_cols = self._cons_addnl_fields(
            columns, addnl_fields)

        if prefixlen and ('prefixlen' not in columns or columns != ['*']):
            addnl_fields.append('prefixlen')
            drop_cols.append('prefixlen')

        # /32 routes are stored with the /32 prefix, so if user doesn't specify
        # prefix as some folks do, assume /32
        newpfx = []
        for item in prefix:
            ipvers = self._get_ipvers(item)

            if item and '/' not in item:
                if ipvers == 4:
                    item += '/32'
                else:
                    item += '/128'

            newpfx.append(item)

        df = super().get(addnl_fields=addnl_fields, prefix=newpfx,
                         ipvers=ipvers, **kwargs)

        if not df.empty and 'prefix' in df.columns:
            df = df.loc[df['prefix'] != "127.0.0.0/8"]
            df['prefix'].replace('default', '0.0.0.0/0', inplace=True)

            if prefixlen or 'prefixlen' in columns or (columns == ['*']):
                # This convoluted logic to handle the issue of invalid entries
                # for prefix in JUNOS routing table
                df['prefixlen'] = df['prefix'].str.split('/')
                df = df[df.prefixlen.str.len() == 2]
                df['prefixlen'] = df['prefixlen'].str[1].astype('int')

            if prefixlen:
                if any(map(prefixlen.startswith, ['<', '>'])):
                    query_str = f'prefixlen {prefixlen}'
                elif prefixlen.startswith('!'):
                    query_str = f'prefixlen != {prefixlen[1:]}'
                else:
                    query_str = f'prefixlen == {prefixlen}'

                # drop in reset_index to not add an additional index col
                df = df.query(query_str).reset_index(drop=True)

            if columns != ['*'] and 'prefixlen' not in columns:
                drop_cols.append('prefixlen')

        if not df.empty and ('numNexthops' in columns or (columns == ['*'])):
            srs_oif = df['oifs'].str.len()
            srs_hops = df['nexthopIps'].str.len()
            srs = np.array(list(zip(srs_oif, srs_hops)))
            srs_max = np.amax(srs, 1)
            df.insert(len(df.columns)-1, 'numNexthops', srs_max)

        if user_query:
            df = df.query(user_query).reset_index(drop=True)
        if drop_cols:
            df.drop(columns=drop_cols, inplace=True, errors='ignore')

        return df

    def summarize(self, **kwargs):

        self._init_summarize(self.iobj._table, **kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('uniquePrefixCnt', 'prefix', 'nunique'),
            ('uniqueVrfsCnt', 'vrf', 'nunique'),
        ]

        self._summarize_on_perdevice_stat = [
            ('routesPerHostStat', '', 'prefix', 'count')
        ]

        self._summarize_on_add_with_query = [
            ('ifRoutesCnt',
             'prefixlen == 30 or prefixlen == 31', 'prefix'),
            ('hostRoutesCnt', 'prefixlen == 32', 'prefix'),
            ('totalV4RoutesinNs', 'ipvers == 4', 'prefix'),
            ('totalV6RoutesinNs', 'ipvers == 6', 'prefix'),
        ]

        self._summarize_on_add_list_or_count = [
            ('routingProtocolCnt', 'protocol'),
            ('nexthopCnt', 'numNexthops'),
        ]

        self._gen_summarize_data()

        # Now for the stuff that is specific to routes
        routes_per_vrfns = self.summary_df.groupby(by=["namespace", "vrf"])[
            "prefix"].count().groupby("namespace")
        self._add_stats_to_summary(routes_per_vrfns, 'routesperVrfStat')
        self.summary_row_order.append('routesperVrfStat')

        device_with_defrt_per_vrfns = self.summary_df \
            .query('prefix == "0.0.0.0/0"') \
            .groupby(by=["namespace", "vrf"])[
                "hostname"].nunique()
        devices_per_vrfns = self.summary_df.groupby(by=["namespace", "vrf"])[
            "hostname"].nunique()

        {self.ns[i[0]].update({
            "deviceWithNoDefRoute":
            device_with_defrt_per_vrfns[i] == devices_per_vrfns[i]})
         for i in device_with_defrt_per_vrfns.keys() if i[0] in self.ns.keys()}
        self.summary_row_order.append('deviceWithNoDefRoute')

        self._post_summarize()
        return self.ns_df.convert_dtypes()

    def lpm(self, **kwargs):
        if not self.iobj._table:
            raise NotImplementedError

        drop_cols = []

        addr = kwargs.pop('address')
        kwargs.pop('ipvers', None)
        df = kwargs.pop('cached_df', pd.DataFrame())
        addnl_fields = kwargs.pop('addnl_fields', [])

        try:
            ipaddr = ip_address(addr)
            ipvers = ipaddr._version
        except ValueError as e:
            raise ValueError(e)

        cols = kwargs.pop("columns", ["namespace", "hostname", "vrf", "metric",
                                      "prefix", "prefixlen", "nexthopIps",
                                      "oifs", "protocol", "ipvers"])

        if cols != ['default'] and cols != ['*']:
            if 'prefix' not in cols:
                addnl_fields.insert(-1, 'prefix')
                drop_cols.append('prefix')

        rslt = pd.DataFrame()

        # if not using a pre-populated dataframe
        if df.empty:
            df = self.get(ipvers=ipvers, columns=cols,
                          addnl_fields=addnl_fields, **kwargs)
        else:
            df = df.query(f'ipvers=={ipvers}')

        if df.empty:
            return df

        if 'prefixlen' not in df.columns:
            df['prefixlen'] = df['prefix'].str.split('/')
            df = df[df.prefixlen.str.len() == 2]
            df['prefixlen'] = df['prefixlen'].str[1].astype('int')
            drop_cols.append('prefixlen')

        # Vectorized operation for faster results with IPv4:
        if ipvers == 4:
            intaddr = df.prefix.str.split('/').str[0] \
                .map(lambda y: int(''.join(['%02x' % int(x)
                                            for x in y.split('.')]),
                                   16))
            netmask = df.prefixlen \
                .map(lambda x: (0xffffffff << (32 - x)) & 0xffffffff)
            match = (ipaddr._ip & netmask) == (intaddr & netmask)
            rslt = df.loc[match.loc[match].index] \
                .sort_values('prefixlen', ascending=False) \
                .drop_duplicates(['namespace', 'hostname', 'vrf'])
        else:
            selected_entries = {}
            max_plens = defaultdict(int)
            for row in df.itertuples():
                rtentry = ip_network(row.prefix)
                if ipaddr in rtentry:
                    key = f'{row.namespace}-{row.hostname}-{row.vrf}'
                    if rtentry.prefixlen > max_plens[key]:
                        max_plens[key] = rtentry.prefixlen
                        selected_entries[key] = row
            if selected_entries:
                rslt = pd.DataFrame(list(selected_entries.values())) \
                    .drop(columns=['Index'], errors='ignore')
            else:
                rslt = pd.DataFrame()

        if drop_cols:
            return rslt.drop(columns=drop_cols, errors='ignore')
        else:
            return rslt

    def aver(self, **kwargs) -> pd.DataFrame:
        """Verify that the routing table is consistent
        The only check for now is to ensure every host has a default route/vrf'
        """
        df = self.get(**kwargs)
        if df.empty:
            return df



import time
from nubia import command, argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.evpnVni import EvpnvniObj


@command("evpnVni", help="Act on EVPN VNI data")
class EvpnVniCmd(SqCommand):
    """EVPN information such as VNI/VLAN mapping, VTEP IPs etc."""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        query_str: str = " ",
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=EvpnvniObj,
        )

    @command("show")
    @argument("vni", description="VNI ID to qualify")
    @argument("priVtepIp", description="Primary VTEP IP to qualify")
    def show(self, vni: str = "", priVtepIp: str = ''):
        """Show EVPN VNI info
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname,
                                vni=vni.split(),
                                priVtepIp=priVtepIp.split(),
                                columns=self.columns,
                                query_str=self.query_str,
                                namespace=self.namespace,
                                )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @command("assert")
    @argument("status", description="Show only assert that matches this value",
              choices=["all", "fail", "pass"])
    def aver(self, status: str = 'pass'):
        """Assert BGP is functioning properly"""

        now = time.time()

        df = self._invoke_sqobj(self.sqobj.aver,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                status=status,
                                )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._assert_gen_output(df)
